//---------------------------------------------------------------------------
/*
Remote Train Control Program for Windows

 Copyright 2019 by Mark DiVecchio

    This file is part of Remote Train Control.

    Remote Train Control is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Remote Train Control is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

*/
// Home Page : http://www.silogic.com/trains/RTC_Running.html
//---------------------------------------------------------------------------
#ifndef Comm_ThreadH
#define Comm_ThreadH
//---------------------------------------------------------------------------
#include <Classes.hpp>
//---------------------------------------------------------------------------
#define DWDELAY 60000		// default timeout of 60 seconds on some WaitForSingleEvent calls

// Types of commands in the FIFO and Dispatch Lists
//
#define CMD_Type_SendString 0
#define CMD_Type_WaitForResponse 1
#define CMD_Type_SendHex 2
#define CMD_Type_DelayedCommand 3
//---------------------------------------------------------------------------
//
// Convert a RAM address (0 - 0x3FF) and byte number (0-3) into a "q" command address
//
#define RAMTOQ(address, byteno) ( ((byteno) << 14) | (((address) & 0x0001) << 13 | ((address) & 0xFFE) >> 1) ^ 0x1F80)
//---------------------------------------------------------------------------
/*
Message for communicating between the Comm Thread and the Main VCL UI Thread
*/
struct UI_Struct {
	String Line;
    TColor PrintColor;
    TAbNumEdit * Countr;
    TAbNumSpin * Spinner;
    TColor ColorDefault;
    TEdit * EditBox;
    TRichEdit * Buffer;
    TAbLED * LED;
    TComboBox * ComboBox;
    union {
	    TStringGrid * StringGrid;
        TPanel * Panel;
        };
    TLabel * Label;
    union {
	    int iValue;
        int iSignalNo;
        };
    int * pValue;
	union {
    	int iTIU;
        int iNumberPlate;
        };
    float fValue;
    bool bValue;
    union {
    	int iChan;
        int iNumLights;
        };
	union {
	    int iState;
        int iAspect;
        int iDistant;
        };
    union {
	    int iAIU;
        int iCol;
        };
    union {
	    int PCNo;
        int iRow;
        };
    int iDir;
    int iType;
    String Status;   // Message 28
    String Caption;  // Message 28
    };

class TCommThread : public TThread
{
private:
	__fastcall TCommThread(bool CreateSuspended);
    int * MyResultValue;
    bool __fastcall TIUSendWaitforResponseRetry(unsigned char ConsistNo, byte * Command,
		int ExpectedCount, int WaitTime, unsigned char Remote, unsigned char TIU, int * ptrRespCode,
    	int * SWResponseCount, byte * SWResponseBuffer);
    bool __fastcall Send_String(unsigned char Consist, byte * Command_String,
		unsigned char TIU, unsigned char Remote, int Expected, int WaitTime, bool Record_Flag, byte * ptrLashUpEngineList,
    	int * ptrRespCode, int * SSResponseCount, byte * SSResponseBuffer);
    bool __fastcall SendHexStr(byte * Cmd_In, int Len, int ExpectedByteCount, int Timeout, char cTIULocal);

    unsigned int __fastcall CRCGen(Cmd * Command, int StartChar, int EndChar, bool Quiet);
    unsigned int __fastcall CRCEvaluation(Cmd * Cmd_String, int StartChar, int EndChar, bool Quiet);

    int NumRetries;
    bool Speed_Mode, TXOnlyComm, Echo;
	TForm * MyCaller;
	HWND ParentHandle;

    bool My_Radio_COM_Mode, My_Radio_WiFi_Mode, My_RFID_COM_Mode, My_RFID_WiFi_Mode, My_AIU_WiFi_Mode;

    void __fastcall SetName(DWORD dwThreadID, LPCSTR szThreadName);

    bool __fastcall SendReadRAMCommand(struct FIFOItem * PItem);
	void SetName();

    // Radio Send socket
    void __fastcall RadioSendChangeState(TObject *Sender,
          TSocketState OldState, TSocketState NewState);
    void __fastcall RadioSendSendData(TObject *Sender,
      int BytesSent);
	void __fastcall RadioSendDataSent(TObject *Sender, WORD ErrCode);
	void __fastcall RadioSendSessionClosed(TObject *Sender, WORD ErrCode);
   	void __fastcall RadioSendSessionConnected(TObject *Sender, WORD ErrCode);
    // Radio Receive socket
    void __fastcall RadioRecvDataAvailable(TObject *Sender,
      WORD ErrCode);
	void __fastcall RadioRecvChangeState(TObject *Sender,
          TSocketState OldState, TSocketState NewState);
	void __fastcall RadioRecvSessionClosed(TObject *Sender, WORD ErrCode);
	void __fastcall RadioRecvSessionConnected(TObject *Sender, WORD ErrCode);

    String RadioIP;		// IP of WiFi Radio for TIU Communications
    String RadioPort;	// Port for TIU Communications
	bool MultiCast;		// set true for Radio MultiCast TX

    // AIU Send socket
    void __fastcall AIUSendChangeState(TObject *Sender,
          TSocketState OldState, TSocketState NewState);
    void __fastcall AIUSendSendData(TObject *Sender,
      int BytesSent);
	void __fastcall AIUSendDataSent(TObject *Sender, WORD ErrCode);
	void __fastcall AIUSendSessionClosed(TObject *Sender, WORD ErrCode);
   	void __fastcall AIUSendSessionConnected(TObject *Sender, WORD ErrCode);
    String AIUIP;		// IP of AIU WiFi
    String AIUPort;	    // Port for AIU
	bool AIUMultiCast;	// set true for AIU MultiCast TX

    #if __BORLANDC__ >= 0x720
	// Defined for CB10.1 C++ programs that use the VCL.
  	Overbyteicswinsock::TInAddr FServerAddr;
	Overbyteicswinsock::TSockAddrIn Detector;
    #else
    TInAddr FServerAddr;
	TSockAddrIn Detector;
	#endif

	bool RadioWiFiRXConnected, RadioWiFiTXConnected, AIUWiFiTXConnected;

protected:
	void __fastcall Execute();
    void __fastcall CleanUp();
    void __fastcall Initialize();
    void __fastcall RunOnTerminate(TObject * Sender);

public:
    __fastcall TCommThread(bool CreateSuspended, TForm * Caller, HWND Parent, String PortName, int BaudRate, int ParitySel, char BitsSel, int StopSel,
    				bool Radio_COM_Mode, bool Radio_WiFi_Mode, bool RFID_COM_Mode, bool RFID_WiFi_Mode, bool AIU_WiFi_Mode, int * ResultValue);
    __fastcall ~TCommThread();

    bool __fastcall Send_String_FIFO(unsigned char Consist, byte * Command_String,
		unsigned char TIU, unsigned char Remote, int Expected, int WaitTime, bool Record_Flag, byte * ptrLashUpEngineList = (byte *) NULL,
		int * ptrRespCode = (int *) NULL, int * SSResponseCount = (int *) NULL, byte * SSResponseBuffer = (byte *) NULL);

	bool __fastcall TIUSendWaitforResponse_FIFO(unsigned char Consist, byte * Command,
		int ExpectedCount, int WaitTime, unsigned char Remote, unsigned char TIU, int * ptrRespCode = (int *) NULL,
	    int * SWResponseCount = (int *) NULL, byte * SWResponseBuffer = (byte *) NULL);

    bool __fastcall SendHexStrStr_FIFO(int Port, String Cmd_In, int Len);
    bool __fastcall SendHexStr_FIFO(int Port, byte * Cmd_In, int Len);

    void __fastcall charstringtobinary(char * Command, char * Packet, int Len);

    // Thread to VCL communications
    void __fastcall PrintItem(String PrintLine, TColor PrintColor = clBlack);
    void __fastcall Bump(TAbNumEdit * Countr);
    void __fastcall Dec(TAbNumEdit * Countr);
    void __fastcall GetCounterValue(TAbNumEdit * Countr, int * CntrVal);
    void __fastcall SetColor(TAbNumSpin * Spinner, TColor SpinnerColorDefault);
    void __fastcall SetTextValue(TEdit * EditBox, String TextValue);
    void __fastcall GetTextValue(TEdit * EditBox, String &TextValue);
    void __fastcall SetRecordingTextValue(String TextValue);
    void __fastcall ShowinTIUBuffer(String TextValue, TColor Color = clLime);
    void __fastcall PopupItem(String PrintLine, int Duration = 2000);
    void __fastcall ShowinBuffer(TRichEdit * Buf, String PrintLine, TColor PrintColor = clBlack, bool Scroll = true);
    void __fastcall SetNumSpinValue(TAbNumSpin * Spinner, int Value = 0);
	void __fastcall SetNumEditValue(TAbNumEdit * Countr, int Value = 0);
    void __fastcall SetPCLED(TAbLED * LED, int iState, String Line);
    void __fastcall SetNumEditColor(TAbNumEdit * Countr, TColor Color);
    void __fastcall GetComboBoxText(TComboBox * ComboBox, AnsiString &TextValue, int Index);
    void __fastcall ClearCounter(TAbNumEdit * Countr);
    void __fastcall SetCounter(TAbNumEdit * Countr, float value);
    void __fastcall OpenAIUWindow(int TIU, int AIU);
    void __fastcall GetAIUSwitchState(int iChannel, int iAIU, int iTIU, bool &bState, int MyPCNo);
    void __fastcall GetAIUAccState(int iChannel, int iAIU, int iTIU, bool &bState, int MyPCNo);
	void __fastcall GetAIUSwitchName(int iChannel, int iAIU, int iTIU, AnsiString &TextValue, int MyPCNo);
	void __fastcall GetAIUAccName(int iChannel, int iAIU, int iTIU, AnsiString &TextValue, int MyPCNo);
    void __fastcall MakeItSo(int iAIU, int iTIU, int MyPCNo);
	void __fastcall SetVirtualAIUAccState(bool bState, int iChannel, int iAIU, int iTIU, int MyPCNo);
    void __fastcall RTSSwitch(bool Checked);
    void __fastcall DTRSwitch(bool Checked);
	void __fastcall AddCntr(TAbNumEdit * Countr, int Val);

    void __fastcall GetStringGridText(TStringGrid * StringGrid, AnsiString &TextValue, int &PSLevel, int Index);
	void __fastcall SetLabelText(TLabel * Label, String TextValue);
	void __fastcall PopupMessage(String Message, String Caption, String Status, int dly);

	// Layout Flags
	void __fastcall GetFlag(int FlagNo, int &iState);
	void __fastcall SetFlag(int FlagNo, int &iState);
	void __fastcall GetFlagName(int FlagNo, String &FlagName);
	void __fastcall SetFlagName(int FlagNo, String FlagName, int &iState);
    void __fastcall ShowLayout(void);

    // Fixed Signals
	void __fastcall GetAspect(int SignalNo, int &Aspect, int &iNumLights, int &iType);
	void __fastcall SetAspect(int SignalNo, int &Aspect);
	void __fastcall GetSignalName(int SignalNo, String &SignalName, int &NumberPlate, int &iType);
	void __fastcall CreateSignal(int SignalNo, int iDir, int iType, int Col, int Row, String SignalName, int &Aspect, int iNumLights = 0, int NumberPlate = 0);
	void __fastcall DeleteSignal(int SignalNo);
    void __fastcall DeleteAllSignals(void);
    void __fastcall ShowSignals(void);

    // Cab Signals
	void __fastcall SetCabAspect(int Head, int Aspect, int MyPCNo);
    void __fastcall GetCabAspect(int Head, int &iAspect, int &iNumLights, int &iType, int MyPCNo);
	void __fastcall SetCabLabels(int Head, String sName, int iNumberPlate, int iType, int MyPCNo);
    void __fastcall CreateCabSignals(int iHead, int iDistant, int MyPCNo);

    // LED
	void __fastcall SetLED(TAbLED * LED, TColor ColorOn = clLime, bool Checked = true, bool Flashing = false, String Line = NULL);

    bool Finished;	// true when finished

    TThreadList * FIFO;	    //  FIFO Communication List - Thread Safe    

};
//---------------------------------------------------------------------------
#endif

