//---------------------------------------------------------------------------
/*
Remote Train Control Program for Windows

 Copyright 2019 by Mark DiVecchio

    This file is part of Remote Train Control.

    Remote Train Control is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Remote Train Control is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

*/
// Home Page : http://www.silogic.com/trains/RTC_Running.html
//---------------------------------------------------------------------------
#include <vcl.h>
#include <tchar.h>
#pragma hdrstop

#include "Message_Popup.h"
#include "Train_Control.h"
#include "Setup_Control.h"
#include "PBFileThread.h"

#include "Debug_Controls.h"
#include "Super_TIU_Mode.h"
#include "Clipbrd.hpp"
#include "Message_Thread.h"

#if __BORLANDC__ >= 0x720
// Defined for CB10.1 C++ programs that use the VCL.
//#include <stdlib.h>
//#include <stdarg.h>
//#include <stdint.h>
//#include <io.h>
//typedef unsigned int socklen_t;
//#include <stddef.h>
//#include <stdio.h>
//#include <string.h>
//#include <sys/types.h>
//#include <winsock2.h>
#include <ws2ipdef.h>
//#include <process.h>
#else
typedef unsigned short uint16_t;
typedef unsigned int uint32_t;
typedef unsigned int socklen_t;
#include <stdio.h>
//#include <winsock.h> 		// was winsock2.h
//#include <winsock2.h>
#endif

// mDNSThread does mDNS requests searching for mDNS servers
#include "mDNSThread.h"
// mDNSServices does mDNS requests searching for mDNS servers AND responds to mDNS requests
#include "mDNSServices.h"

#include <stdio.h>
#include <stdlib.h>
//---------------------------------------------------------------------------
//
// This form must be auto create because information is retained
// only by this form for the communications parameters
//
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma link "QuitButton"
#pragma link "_GClass"
#pragma link "SyncListBox"
#pragma resource "*.dfm"
TSetup *Setup;

extern bool Ignore_FF_Char;	// used by Comm_Thread.cpp
extern TThread * PBThread;	// pointer to Setup's playback thread
extern struct AIU_EMU AIUEmulator[TIUMAX];				// Flags to indicate an AIU Emulator is attached to an emulated TIU
extern struct MSG_DISPLAY MessageDisplay[100];          // Flags to indicate a message display ( 1 - 100 ) is available
TMultiDNS * ptrMultiDNSThread;				// pointer to MultiDNS thread
mDNSServices * ptrmDNSServices;				// pointer to mDNS Services advertiser
bool mDNSTerminate;							// mDNS Server Terminate flag
//---------------------------------------------------------------------------
__fastcall TSetup::TSetup(TComponent* Owner)
	: TForm(Owner)
{
assert(false);	// never called since I used the overloaded constructor below
}

//---------------------------------------------------------------------------
__fastcall TSetup::TSetup(TComponent* Owner, int TIU, int Remote)
	: TForm(Owner)
{
// grab the TIU & Remote Numbers
TIUNo = TIU;
assert(TIUNo >= 1 && TIUNo <= TIUMAX);
RemoteNo = Remote;
assert(RemoteNo >= 0 && RemoteNo <= REMOTEMAX);

ptrmDNSServices = NULL;		// mDNS server not running

// read preferences from the Registry
RegFileRootKey = MainForm->RegFileRootKey;
// read Preferences from registry if they exist
TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
assert(RegFile != NULL);
__try {
    ModeSelect->ItemIndex = RegFile->ReadInteger("Options", "Mode", ModeSelect->ItemIndex);
	if (ModeSelect->ItemIndex != 1) {	// modes 0 and 2
    	// Wired Mode
		// Saved on a per TIU basis
    	PortNum->ValueAsInt = RegFile->ReadInteger("Speed", "Port" + String(TIUNo), PortNum->ValueAsInt);
    	RFIDNum->ValueAsInt = RegFile->ReadInteger("Speed", "RFID" + String(TIUNo), RFIDNum->ValueAsInt);
		} else { 	// mode 1
        // Radio Mode
	    // Saved on a per program basis
	    PortNum->ValueAsInt = RegFile->ReadInteger("Speed", "Port", PortNum->ValueAsInt);
	    RFIDNum->ValueAsInt = RegFile->ReadInteger("Speed", "RFID", RFIDNum->ValueAsInt);
        }
	// Load the HostIP IP Addresses and Ports
	const String HostIPFile = String(MainForm->cStartDir)  + "\\" + "HostIP.ini";
    IPComboBox->Clear();
	if (FileExists(HostIPFile)) {
        IPComboBox->Clear();
		IPComboBox->Items->LoadFromFile(HostIPFile);
    	// Host IP -  stored by C++ version
        IPComboBox->ItemIndex = RegFile->ReadInteger("Options", "HostIPIndex" + IntToHex(__BORLANDC__, 3), IPComboBox->ItemIndex);
		if (IPComboBox->ItemIndex == -1 && IPComboBox->Items->Count > 0) {
			IPComboBox->ItemIndex = 0;
            } else {
            if (IPComboBox->ItemIndex >= IPComboBox->Items->Count) {
                IPComboBox->ItemIndex = 0;
                }
            }
	    }
    // Saved on a per program basis
    BaudRateSel->ItemIndex = RegFile->ReadInteger("Speed", "Baud", BaudRateSel->ItemIndex);
    ParitySel->ItemIndex = RegFile->ReadInteger("Speed", "Parity", ParitySel->ItemIndex);
    BitsSel->ItemIndex = RegFile->ReadInteger("Speed", "Bits", BitsSel->ItemIndex);
    StopSel->ItemIndex = RegFile->ReadInteger("Speed", "Stop", StopSel->ItemIndex);
	Echo->Checked = RegFile->ReadInteger("Options", "Echo", Echo->Checked);
	Ignore_FF->Checked = RegFile->ReadInteger("Options", "Ignore_FF", Ignore_FF->Checked);
	Hex->Checked = RegFile->ReadInteger("Options", "Hex", Hex->Checked);
	Duplex->Checked = RegFile->ReadInteger("Options", "Duplex", Duplex->Checked);
	RFID_COM_Mode->Checked = RegFile->ReadInteger("Options", "RFID", RFID_COM_Mode->Checked);
	RFID_WiFi_Mode->Checked = RegFile->ReadInteger("Options", "RFIDWiFi", RFID_WiFi_Mode->Checked);
        MainForm->ConnectedRFIDLED->Visible = (RFID_COM_Mode->Checked || RFID_WiFi_Mode->Checked);
	Radio_COM_Mode->Checked = RegFile->ReadInteger("Options", "Radio", Radio_COM_Mode->Checked);
    	Radio_WiFi_Mode->Checked = !Radio_COM_Mode->Checked;		// one or the other can be selected, not both
	AIU_WiFi_Mode->Checked = RegFile->ReadInteger("Options", "AIUWiFi", AIU_WiFi_Mode->Checked);
		AIU_WiFi_ModeClick(NULL);

	AIUIP->Text = RegFile->ReadString("Options", "AIUIP", AIUIP->Text); 	// Should be multicast   239.255.200.3
	AIU_WiFi_Multicast->Checked = RegFile->ReadInteger("Options", "AIU_Multicast", AIU_WiFi_Multicast->Checked);
	AIUPort->Text = RegFile->ReadString("Options", "AIUPort", AIUPort->Text);

	RadioIP->Text = RegFile->ReadString("Options", "RadioIP", RadioIP->Text); 	// Should be a single IP address but
	Radio_WiFi_Multicast->Checked = RegFile->ReadInteger("Options", "Multicast", Radio_WiFi_Multicast->Checked);
																			    // can be Multicast 239.255.200.3
	RadioPort->Text = RegFile->ReadString("Options", "RadioPort", RadioPort->Text);
    	RadioWiFiLabel->Enabled = Radio_WiFi_Mode->Checked;
        Radio_WiFi_Multicast->Enabled = Radio_WiFi_Mode->Checked;
    	RadioIP->Enabled = Radio_WiFi_Mode->Checked;
    	RadioPort->Enabled = Radio_WiFi_Mode->Checked;
    	RFIDWiFiLabel->Enabled = RFID_WiFi_Mode->Checked;
		PortNum->Enabled = Radio_COM_Mode->Checked;
        COMLabel->Enabled = Radio_COM_Mode->Checked;
		RFIDNum->Enabled = RFID_COM_Mode->Checked;
        RFIDLabel->Enabled = RFID_COM_Mode->Checked;
	SpareCheckbox->Checked = RegFile->ReadInteger("Options", "SpareCheckbox", SpareCheckbox->Checked);
  	NumRetries->Value = RegFile->ReadInteger("Options", "NumRetries", NumRetries->Value);
    if (NumRetries->Value > 1 ) NumRetries->Value = 1;	// on startup, NumRetries can't be higher than 1
    													// I've found that additional retries don't help
    // Other setup
  	AllowZero->Checked = RegFile->ReadInteger("Options", "AllowZero", AllowZero->Checked);
  	BrakeBoost->ItemIndex = RegFile->ReadInteger("Options", "BrakeBoost", BrakeBoost->ItemIndex);
  	AutoCoupler->Checked = RegFile->ReadInteger("Options", "AutoCoupler", AutoCoupler->Checked);
  	MaxSpeed->Value = RegFile->ReadInteger("Options", "MaxSpeed", MaxSpeed->Value);
  	EncodedRecordFormat->Checked = RegFile->ReadInteger("Options", "EncodedRecordFormat", EncodedRecordFormat->Checked);
    MainForm->EncodedRecordFormat = EncodedRecordFormat->Checked;
	MPHScaling->Value = RegFile->ReadInteger("Options", "MPHScaling", MPHScaling->Value);
	MainForm->GlobalMPHScaling = MPHScaling->Value;
  	AutoBell->Checked = RegFile->ReadInteger("Options", "AutoBell", AutoBell->Checked);
  	AutoBellOff->Checked = RegFile->ReadInteger("Options", "AutoBellOff", AutoBellOff->Checked);
  	AutoWhistle->Checked = RegFile->ReadInteger("Options", "AutoWhistle", AutoWhistle->Checked);
  	ExtendedSUSD->Checked = RegFile->ReadInteger("Options", "ExtendedSUSD", ExtendedSUSD->Checked);
  	RepeatPlayback->Checked = RegFile->ReadInteger("Options", "RepeatPlayback", RepeatPlayback->Checked);
  	Speed_Mode->Checked = RegFile->ReadInteger("Options", "Speed_Mode", Speed_Mode->Checked);
  	TXOnlyComm->Checked = RegFile->ReadInteger("Options", "TXOnlyComm", TXOnlyComm->Checked);
  	DropDTRonIdle->Checked = RegFile->ReadInteger("Options", "DropDTRonIdle", DropDTRonIdle->Checked);
  	DropRTSonIdle->Checked = RegFile->ReadInteger("Options", "DropRTSonIdle", DropRTSonIdle->Checked);
    DTRAlwaysOn->Checked = RegFile->ReadInteger("Options", "DTRAlwaysOn", DTRAlwaysOn->Checked);
  	KeepEngines->Checked = RegFile->ReadInteger("Options", "KeepEngines", KeepEngines->Checked);
  	StandardAccDec->Checked = RegFile->ReadInteger("Options", "StandardAccDec", StandardAccDec->Checked);
  	AutoConnect->Checked = RegFile->ReadInteger("Options", "AutoConnect", AutoConnect->Checked);
  	ShortBellWhistleSequence->Checked = RegFile->ReadInteger("Options", "ShortBellWhistleSequence", ShortBellWhistleSequence->Checked);
  	AutoPFA->Checked = RegFile->ReadInteger("Options", "AutoPFA", AutoPFA->Checked);
  	DisableEmergencyStopButton->Checked = RegFile->ReadInteger("Options", "DisableEmergencyStop", DisableEmergencyStopButton->Checked);
  	AutoHotButtons->Checked = RegFile->ReadInteger("Options", "AutoHotButtons", AutoHotButtons->Checked);
  	AutoQuickControls->Checked = RegFile->ReadInteger("Options", "AutoQuickControls", AutoQuickControls->Checked);
  	Allow255Sounds->Checked = RegFile->ReadInteger("Options", "Allow255Sounds", Allow255Sounds->Checked);
  	Safety_Mode->Checked = RegFile->ReadInteger("Options", "Safety_Mode", Safety_Mode->Checked);
  	GeneralSounds->Checked = RegFile->ReadInteger("Options", "GeneralSounds", GeneralSounds->Checked);
  	NoSleep->Checked = RegFile->ReadInteger("Options", "NoSleep", NoSleep->Checked);
  	AutoSoundCtrls->Checked = RegFile->ReadInteger("Options", "AutoSoundCtrls", AutoSoundCtrls->Checked);
  	SmokeOFFatStartup->Checked = RegFile->ReadInteger("Options", "SmokeOFF", SmokeOFFatStartup->Checked);
  	Quick_Ctrl_Scaling->Value = RegFile->ReadInteger("Options", "Quick_Ctrl_Scaling", Quick_Ctrl_Scaling->Value * 10) / 10.0;
    TransparentSpin->Value = RegFile->ReadInteger("Options", "Transparent", TransparentSpin->Value);
    // Super TIU mode
  	SuperTIUMode = RegFile->ReadInteger("Options", "SuperTIUMode", SuperTIUMode);
  	//TIUSuperMode[0] = RegFile->ReadInteger("Options", "TIU1SuperMode", TIUSuperMode[0]);
  	//TIUSuperMode[1] = RegFile->ReadInteger("Options", "TIU2SuperMode", TIUSuperMode[1]);
  	//TIUSuperMode[2] = RegFile->ReadInteger("Options", "TIU3SuperMode", TIUSuperMode[2]);
  	//TIUSuperMode[3] = RegFile->ReadInteger("Options", "TIU4SuperMode", TIUSuperMode[3]);
  	//TIUSuperMode[4] = RegFile->ReadInteger("Options", "TIU5SuperMode", TIUSuperMode[4]);
    for (int iTIU = 1 ; iTIU <= TIUMAX ; iTIU++) {	 // TIU 1-5
    	// Super TIU Mode by TIU number
	  	TIUSuperMode[iTIU-1] = RegFile->ReadInteger("Options", "TIU" + String(iTIU) + "SuperMode", TIUSuperMode[iTIU-1]);
        // AIU Emulator Attached to this TIU number
        //AIUEmulator[iTIU-1].Found = RegFile->ReadInteger("Options", "TIU" + String(iTIU) + "AIUEmulator", AIUEmulator[iTIU-1].Found);
    	}
	}
__finally {
	// Deallocate RegFile
	delete RegFile;
	}

// Load the Radio WiFi IP Addresses and Ports
const String WiFiRadioFile = String(MainForm->cStartDir)  + "\\" + "WiFiRadio.ini";
if (FileExists(WiFiRadioFile)) {
	WiFiListBox->Clear();
	WiFiListBox->Items->LoadFromFile(WiFiRadioFile);
    }
// Load the AIU WiFi IP Addresses and Ports
const String AIURadioFile = String(MainForm->cStartDir)  + "\\" + "WiFiAIU.ini";
if (FileExists(AIURadioFile)) {
	AIUWiFiListBox->Clear();
	AIUWiFiListBox->Items->LoadFromFile(AIURadioFile);
    }
// in debugging mode, make a few more controls visible
if (DEBUGGING) {
	Radio_WiFi_Multicast->Visible = true;
	AIU_WiFi_Multicast->Visible = true;
    }
//

CommSetupPageControl->ActivePage = SerialPortSetup;

if ((ModeSelect->ItemIndex == 0)) {
	// "Wired" selected
	COMSelGroup->Caption = "Serial Port for TIU" + String(TIUNo);
	} else {
    Set_Comment();
    }
//
// Select a Host IP for your network
if (Radio_WiFi_Mode->Checked || RFID_WiFi_Mode->Checked || AIU_WiFi_Mode->Checked) {
	HostIPPage->Enabled = true;
    // next line moved to  [CONNECT] button on Main Window
   	//StartmDNSServices(RFID_WiFi_Mode->Checked, Radio_WiFi_Mode->Checked);	// Start the mDNS responder if in WiFi mode
    }
//
PBThread = NULL;
DragAcceptFiles (Handle, True);		//  Accept drag and drop from the Explorer
// Sleep
if (NoSleep->Checked) {
    // We could #include windows.h to get these constants
    #define ES_CONTINUOUS 0x80000000
    #define ES_SYSTEM_REQUIRED 0x00000001
    #define ES_AWAYMODE_REQUIRED 0x00000040
    // This sets the appropriate flags to prevent the system going into sleep mode; this is useful if making a screen saver that processes data in the background and can't have the computer sleep
    SetThreadExecutionState(ES_CONTINUOUS | ES_SYSTEM_REQUIRED | ES_AWAYMODE_REQUIRED);
    }
//
//
// Setup window is always opaque
//
}
//---------------------------------------------------------------------------
__fastcall TSetup::~TSetup(void)
{
if (PBThread != NULL) {
	PBThread->Terminate();
	PBThread->WaitFor();	// do not use FreeOnTerminate
	delete PBThread;
	PBThread = NULL;	// release the pointer
	}
HideButtonClick(HideButton);   // save setting to Registry
// alternate coding : HideButton->OnClick(HideButton);	// save setting to Registry
// Stop mDNS Server ////- Now, I let the mDNS server terminate automatically when Setup exits
// next line moved to [DISCONNECT] button on Main Window
// StopmDNSServices(RFID_WiFi_Mode->Checked, Radio_WiFi_Mode->Checked);		// Stop the mDNS responder if its running
if (NoSleep->Checked) {
    // We could #include windows.h to get this constant
    #define ES_CONTINUOUS 0x80000000
	// This allows the system to sleep normally
	SetThreadExecutionState(ES_CONTINUOUS);
    }
}
//---------------------------------------------------------------------------

void __fastcall TSetup::HideButtonClick(TObject *Sender)
{
if (SuperTIU) SuperTIU->Hide();	// hide the Super TIU Mode Setup window

// if we just cleared data, don't save it
if (MainForm->bSaveRegistryInfo) {
    // write preferences to the Registry
    TRegIniFile * const RegFile = new TRegIniFile(RegFileRootKey);
    assert(RegFile != NULL);
    __try {
        RegFile->WriteInteger("Options", "Mode", ModeSelect->ItemIndex);
        if (ModeSelect->ItemIndex != 1) {	// modes 0 and 2
            // Wired Mode
            // Saved on a per TIU basis
            RegFile->WriteInteger("Speed", "Port" + String(TIUNo), PortNum->ValueAsInt);
            RegFile->WriteInteger("Speed", "RFID" + String(TIUNo), RFIDNum->ValueAsInt);
            } else {	// mode 1
            // Radio Mode
            // Saved on a per program basis
            RegFile->WriteInteger("Speed", "Port", PortNum->ValueAsInt);
            RegFile->WriteInteger("Speed", "RFID", RFIDNum->ValueAsInt);
            }
        RegFile->WriteInteger("Speed", "Baud", BaudRateSel->ItemIndex);
        RegFile->WriteInteger("Speed", "Parity", ParitySel->ItemIndex);
        RegFile->WriteInteger("Speed", "Bits", BitsSel->ItemIndex);
        RegFile->WriteInteger("Speed", "Stop", StopSel->ItemIndex);
        RegFile->WriteInteger("Options", "Echo", Echo->Checked);
        RegFile->WriteInteger("Options", "Ignore_FF", Ignore_FF->Checked);
        RegFile->WriteInteger("Options", "Hex", Hex->Checked);
        RegFile->WriteInteger("Options", "Duplex", Duplex->Checked);
        // WiFi setup
        RegFile->WriteInteger("Options", "RFID", RFID_COM_Mode->Checked);
        RegFile->WriteInteger("Options", "RFIDWiFi", RFID_WiFi_Mode->Checked);
        RegFile->WriteInteger("Options", "AIUWiFi", AIU_WiFi_Mode->Checked);
        RegFile->WriteInteger("Options", "Radio", Radio_COM_Mode->Checked);
        RegFile->WriteString("Options", "RadioIP", RadioIP->Text);
        RegFile->WriteString("Options", "RadioPort", RadioPort->Text);
        RegFile->WriteInteger("Options", "Multicast", Radio_WiFi_Multicast->Checked);
        RegFile->WriteString("Options", "AIUIP", AIUIP->Text);
        RegFile->WriteString("Options", "AIUPort", AIUPort->Text);
        RegFile->WriteInteger("Options", "AIU_Multicast", AIU_WiFi_Multicast->Checked);

        RegFile->WriteInteger("Options", "SpareCheckbox", SpareCheckbox->Checked);
        RegFile->WriteInteger("Options", "NumRetries", NumRetries->Value);
		// Host IP -  stored by C++ version
        RegFile->WriteInteger("Options", "HostIPIndex" + IntToHex(__BORLANDC__, 3), IPComboBox->ItemIndex);
		// Save the HostIP IP Addresses
		const String HostIPFile = String(MainForm->cStartDir)  + "\\" + "HostIP.ini";
        IPComboBox->Items->SaveToFile(HostIPFile);
        // other setup
        RegFile->WriteInteger("Options", "AllowZero", AllowZero->Checked);
        RegFile->WriteInteger("Options", "BrakeBoost", BrakeBoost->ItemIndex);
        RegFile->WriteInteger("Options", "AutoCoupler", AutoCoupler->Checked);
        RegFile->WriteInteger("Options", "MaxSpeed", MaxSpeed->Value);
        RegFile->WriteInteger("Options", "EncodedRecordFormat", EncodedRecordFormat->Checked);
        RegFile->WriteInteger("Options", "MPHScaling", MPHScaling->Value);
        RegFile->WriteInteger("Options", "AutoBell", AutoBell->Checked);
        RegFile->WriteInteger("Options", "AutoBellOff", AutoBellOff->Checked);
        RegFile->WriteInteger("Options", "AutoWhistle", AutoWhistle->Checked);
        RegFile->WriteInteger("Options", "ExtendedSUSD", ExtendedSUSD->Checked);
        RegFile->WriteInteger("Options", "RepeatPlayback", RepeatPlayback->Checked);
        RegFile->WriteInteger("Options", "Speed_Mode", Speed_Mode->Checked);
        RegFile->WriteInteger("Options", "TXOnlyComm", TXOnlyComm->Checked);
        RegFile->WriteInteger("Options", "DropDTRonIdle", DropDTRonIdle->Checked);
        RegFile->WriteInteger("Options", "DropRTSonIdle", DropRTSonIdle->Checked);
        RegFile->WriteInteger("Options", "DTRAlwaysOn", DTRAlwaysOn->Checked);
        RegFile->WriteInteger("Options", "KeepEngines", KeepEngines->Checked);
        RegFile->WriteInteger("Options", "StandardAccDec", StandardAccDec->Checked);
        RegFile->WriteInteger("Options", "AutoConnect", AutoConnect->Checked);
        RegFile->WriteInteger("Options", "ShortBellWhistleSequence", ShortBellWhistleSequence->Checked);
        RegFile->WriteInteger("Options", "AutoPFA", AutoPFA->Checked);
        RegFile->WriteInteger("Options", "DisableEmergencyStop", DisableEmergencyStopButton->Checked);
        RegFile->WriteInteger("Options", "AutoHotButtons", AutoHotButtons->Checked);
        RegFile->WriteInteger("Options", "AutoQuickControls", AutoQuickControls->Checked);
        RegFile->WriteInteger("Options", "Safety_Mode", Safety_Mode->Checked);
        RegFile->WriteInteger("Options", "GeneralSounds", GeneralSounds->Checked);
        RegFile->WriteInteger("Options", "NoSleep", NoSleep->Checked);
        RegFile->WriteInteger("Options", "Allow255Sounds", Allow255Sounds->Checked);
        RegFile->WriteInteger("Options", "AutoSoundCtrls", AutoSoundCtrls->Checked);
        RegFile->WriteInteger("Options", "SmokeOFF", SmokeOFFatStartup->Checked);
        RegFile->WriteInteger("Options", "Quick_Ctrl_Scaling", Quick_Ctrl_Scaling->Value * 10);
        RegFile->WriteInteger("Options", "Transparent", TransparentSpin->Value);
        // Super TIU Mode
        RegFile->WriteInteger("Options", "SuperTIUMode", SuperTIUMode);
        //RegFile->WriteInteger("Options", "TIU1SuperMode", TIUSuperMode[0]);
        //RegFile->WriteInteger("Options", "TIU2SuperMode", TIUSuperMode[1]);
        //RegFile->WriteInteger("Options", "TIU3SuperMode", TIUSuperMode[2]);
        //RegFile->WriteInteger("Options", "TIU4SuperMode", TIUSuperMode[3]);
        //RegFile->WriteInteger("Options", "TIU5SuperMode", TIUSuperMode[4]);
    	for (int iTIU = 1 ; iTIU <= TIUMAX ; iTIU++) {	 // TIU 1-5
        	// Super TIU Mode by TIU number
	  		RegFile->WriteInteger("Options", "TIU" + String(iTIU) + "SuperMode", TIUSuperMode[iTIU-1]);
        	// AIU Emulator Attached to this TIU number
        	//RegFile->WriteInteger("Options", "TIU" + String(iTIU) + "AIUEmulator", AIUEmulator[iTIU-1].Found);
    		}

        // Copyright
        RegFile->WriteString("Copyright", "Symbol", "");
        RegFile->WriteString("Copyright", "By", "Mark C. DiVecchio");
        RegFile->WriteString("Copyright", "Year", "2022");
        }
    __finally {
        // Deallocate RegFile
        delete RegFile;
        }
    }
// Save the Radio WiFi IP Addresses and Ports
const String WiFiRadioFile = String(MainForm->cStartDir)  + "\\" + "WiFiRadio.ini";
WiFiListBox->Items->SaveToFile(WiFiRadioFile);
// Save the AIU WiFi IP Addresses and Ports
const String AIURadioFile = String(MainForm->cStartDir)  + "\\" + "WiFiAIU.ini";
AIUWiFiListBox->Items->SaveToFile(AIURadioFile);
Hide();
}
//---------------------------------------------------------------------------

void __fastcall TSetup::AllowZeroClick(TObject *Sender)
{
// This checkbox allows the user to select engine 0, which is sent
// in the command as 1. This should allow you to use engines which
// have been factory reset and not "added" yet
if (AllowZero->Checked) {
	MainForm->EngineSpin->MinValue = 0;   // New Min Value
    } else {
    MainForm->EngineSpin->MinValue = 1;
    // if the Engine number is 0, set it to 1
    if (MainForm->EngineSpin->Value == 0) MainForm->EngineSpin->Value = 1;
    }
}
//---------------------------------------------------------------------------

void __fastcall TSetup::Radio_COM_ModeClick(TObject *Sender)
{
//
// Setup Radio Comm mode
//
// only one radio mode can be selected and one must be selected
Radio_WiFi_Mode->Checked = !Radio_COM_Mode->Checked; // one or the other can be selected, not both
//
// Update labels on MainForm
//
if (Radio_COM_Mode->Checked){

    PortNum->Enabled = true;
    COMLabel->Enabled = true;
    COMCheckListBox->Enabled = true;
    } else {
    PortNum->Enabled = false;
    COMLabel->Enabled = false;
    COMCheckListBox->Enabled = false;
    }
Set_Comment();
}
//---------------------------------------------------------------------------

void __fastcall TSetup::Radio_WiFi_ModeClick(TObject *Sender)
{
//
// Setup Radio WiFi Mode
//
// only one radio mode can be selected and one must be selected
//
Radio_COM_Mode->Checked = !Radio_WiFi_Mode->Checked;	// one or the other can be selected, not both
//
// Update labels on MainForm
//
if (Radio_WiFi_Mode->Checked) {

    // next line moved to  [CONNECT] button on Main Window
    //StartmDNSServices(RFID_WiFi_Mode->Checked, Radio_WiFi_Mode->Checked);		// Start the mDNS responder if in Radio WiFi mode
    HostIPPage->Enabled = true;
    //
    RadioWiFiLabel->Enabled = true;
    WiFiListBox->Enabled = true;
    Radio_WiFi_Multicast->Enabled = true;
    RadioIP->Enabled = true;
    RadioPort->Enabled = true;
	} else {
    RadioWiFiLabel->Enabled = false;
    WiFiListBox->Enabled = false;
    Radio_WiFi_Multicast->Enabled = false;
    RadioIP->Enabled = false;
    RadioPort->Enabled = false;
    }
Set_Comment();
}
//---------------------------------------------------------------------------

void __fastcall TSetup::RFID_COM_ModeClick(TObject *Sender)
{
//
// setup RFID COM Port
//
// none, one or both RFID modes can be selected
//
MainForm->ConnectedRFIDLED->Visible = (RFID_WiFi_Mode->Checked || RFID_COM_Mode->Checked);
//
// Update labels on MainForm
//
if (RFID_COM_Mode->Checked){

    MainForm->RFIDBuffer->Enabled = true;
    MainForm->RFIDBuffer->Color = clWindow;
    //
    RFIDNum->Enabled = true;
    RFIDLabel->Enabled = true;
    } else {
	if (RFID_WiFi_Mode->Checked) {
        MainForm->ConnectedRFIDLED->Caption = "WiFi";
    	}
    MainForm->RFIDBufferLabel->Caption = "RFID Buffer " + MainForm->ConnectedRFIDLED->Caption;
    RFIDLabel->Enabled = false;
    }
Set_Comment();
}
//---------------------------------------------------------------------------

void __fastcall TSetup::RFID_WiFi_ModeClick(TObject *Sender)
{
// Setup RFID WiFi Mode
//
// none, one or both RFID modes can be selected
//
MainForm->ConnectedRFIDLED->Visible = (RFID_WiFi_Mode->Checked || RFID_COM_Mode->Checked);
//
// Update labels on MainForm
//
if (RFID_WiFi_Mode->Checked) {
    MainForm->RFIDBuffer->Enabled = true;
    MainForm->RFIDBuffer->Color = clWindow;
	//
    // next line moved to  [CONNECT] button on Main Window
    //StartmDNSServices(RFID_WiFi_Mode->Checked, Radio_WiFi_Mode->Checked);		// Start the mDNS responder if in RFID WiFi mode
    HostIPPage->Enabled = true;
	//
    RFIDWiFiLabel->Enabled = true;
    } else {
	if (RFID_COM_Mode->Checked) {
        MainForm->ConnectedRFIDLED->Caption = "COM" + String(RFIDNum->ValueAsInt);
    	}
    MainForm->RFIDBufferLabel->Caption = "RFID Buffer " + MainForm->ConnectedRFIDLED->Caption;
	//
    RFIDWiFiLabel->Enabled = false;
    }
Set_Comment();
}
//---------------------------------------------------------------------------

void __fastcall TSetup::Set_Comment()
{
//
// Set title on COM and WiFI select group boxes
//
if (RFID_COM_Mode->Checked && Radio_COM_Mode->Checked)
    COMSelGroup->Caption = "Serial Ports for Radio && RFID Tags";
if (Radio_COM_Mode->Checked && !RFID_COM_Mode->Checked)
    COMSelGroup->Caption = "Serial Port for Radio";
if (!Radio_COM_Mode->Checked && RFID_COM_Mode->Checked)
    COMSelGroup->Caption = "Serial Port for RFID Tags";
if (!Radio_COM_Mode->Checked && !RFID_COM_Mode->Checked)
    COMSelGroup->Caption = "No Serial Ports Used";
if (RFID_WiFi_Mode->Checked && Radio_WiFi_Mode->Checked)
    WiFiSelGroup->Caption = "WiFi Ports for Radio && RFID Tags";
if (Radio_WiFi_Mode->Checked && !RFID_WiFi_Mode->Checked)
    WiFiSelGroup->Caption = "WiFi Port for Radio";
if (!Radio_WiFi_Mode->Checked && RFID_WiFi_Mode->Checked)
    WiFiSelGroup->Caption = "WiFi Port for RFID Tags";
if (!Radio_WiFi_Mode->Checked && !RFID_WiFi_Mode->Checked)
    WiFiSelGroup->Caption = "No WiFi Ports Used";
if (AIU_WiFi_Mode->Checked) {
	AIUWiFiSelGroup->Caption = "WiFi Port for AIU Emulator";
	} else {
	AIUWiFiSelGroup->Caption = "No AIU Emulators";
    }
if (Radio_COM_Mode->Checked) {
	MainForm->ConnectedRadioLED->Caption = "COM" + String(PortNum->ValueAsInt);
	MainForm->TIUBufferLabel->Caption = "TIU Buffer " + MainForm->ConnectedRadioLED->Caption;
	}
if (Radio_WiFi_Mode->Checked) {
	MainForm->ConnectedRadioLED->Caption = "WiFi";
	MainForm->TIUBufferLabel->Caption = "TIU Buffer " + MainForm->ConnectedRadioLED->Caption;
	}
if (RFID_COM_Mode->Checked && RFID_WiFi_Mode->Checked) {
	MainForm->ConnectedRFIDLED->Caption = "WiFi/COM" + String(RFIDNum->ValueAsInt);
	MainForm->RFIDBufferLabel->Caption = "RFID Buffer " + MainForm->ConnectedRFIDLED->Caption;
	} else if (RFID_COM_Mode->Checked) {
	    MainForm->ConnectedRFIDLED->Caption = "COM" + String(RFIDNum->ValueAsInt);
		MainForm->RFIDBufferLabel->Caption = "RFID Buffer " + MainForm->ConnectedRFIDLED->Caption;
    	} else if (RFID_WiFi_Mode->Checked) {
	        MainForm->ConnectedRFIDLED->Caption = "WiFi";
			MainForm->RFIDBufferLabel->Caption = "RFID Buffer " + MainForm->ConnectedRFIDLED->Caption;
        	}
}
//---------------------------------------------------------------------------

void __fastcall TSetup::RecordStartButtonClick(TObject *Sender)
{
// Start Recording the strings sent to the TIU
if (!MainForm->IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	return;
    }
if (MainForm->bPlayingBack) {
    MainForm->DebugMessage("You must stop the previous Playback before you can start a new Recording");
	return;
    }
// prepare to record
MainForm->EncodedRecordFormat = EncodedRecordFormat->Checked;
MainForm->RecordingBuffer->Clear();	// make sure the buffer is empty
MainForm->bRecording = true;
RecordingLED->Checked = true;
RecordingLED->Visible = true;
// set buttons to recording state
RecordStartButton->Enabled = false;
RecordStopButton->Enabled = true;
PlaybackStartButton->Enabled = false;
if (DEBUGGING) DB->Debug_Print("Recording Started");
}
//---------------------------------------------------------------------------

void __fastcall TSetup::RecordStopButtonClick(TObject *Sender)
{
assert(MainForm->bRecording); 	// we must be recording

// recording complete, turn off LED
RecordingLED->Visible = false;

String NameOfFile;
RecordSaveDialog->InitialDir = MainForm->WorkingDir;
RecordSaveDialog->DefaultExt = String("rtc");
RecordSaveDialog->Title = "Save Recording to a File";
RecordSaveDialog->Filter = "Remote Train Control files (*.rtc)|*.rtc";
RecordSaveDialog->FileName = "*.rtc";
RecordSaveDialog->Options = RecordSaveDialog->Options >> ofOverwritePrompt << ofEnableSizing << ofPathMustExist;
													// RTC will handle Overwrite Prompt
if(RecordSaveDialog->Execute()) {
	// user pressed SAVE
	NameOfFile = RecordSaveDialog->FileName;
  	// Change working directory to whatever was selected
   	MainForm->WorkingDir = ExtractFilePath(NameOfFile);
   	// if file exists, confirm first
   	if (FileExists(NameOfFile)) {
   	    // ask for confirmation
   	    int result = Application->MessageBox(TEXT("File Exists - Do you want to overwrite it?"),
   	    TEXT("File Exists"), MB_YESNOCANCEL | MB_ICONQUESTION);
   		// if No, exit
   		if (result  == IDNO) {
        	// end the recording without writing to the file
			MainForm->bRecording = false;
			RecordingLED->Checked = false;
			RecordingLED->Visible = false;
            // set buttons to normal state
			RecordStopButton->Enabled = false;
            RecordStopButton->Enabled = false;
			RecordStartButton->Enabled = true;
			PlaybackStartButton->Enabled = true;
   	        return;		// user refused to overwrite an exisiting file
   	    	}
        if (result == IDCANCEL) {
        	// cancel this dialog and let the recording continue
            RecordingLED->Visible = true;
            return;
            }
   		}
    } else {
    // RecordStopButton is still enabled in case the user wants to try again to save the file
    RecordingLED->Visible = true;
    return;	// user canceled the File Save Dialog
    }
// At this point, it's either a new file or the user has indicated overwrite
MainForm->RecordingBuffer->Lines->SaveToFile(NameOfFile);	// PlainText is set to true in the Object Inspector
MainForm->RecordingBuffer->Clear();	// clear the buffer for next time
// Stop recording
MainForm->bRecording = false;
RecordingLED->Checked = false;
RecordingLED->Visible = false;

// set buttons to normal state
RecordStopButton->Enabled = false;
PlaybackStopButton->Enabled = false;
RecordStartButton->Enabled = true;
PlaybackStartButton->Enabled = true;

//
if (DEBUGGING) DB->Debug_Print("Recording Stopped : " + NameOfFile);
}
//---------------------------------------------------------------------------

void __fastcall TSetup::PlaybackStartButtonClick(TObject *Sender)
{
if (!MainForm->IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	MainForm->bPlayingBack = false;
	return;
    }
if (MainForm->bRecording) {
    MainForm->DebugMessage("You must stop Recording before you can Playback a previous Recording");
	return;
    }
String NameOfFile;
// start playback
PlaybackDialog->InitialDir = MainForm->WorkingDir;
PlaybackDialog->DefaultExt = String("rtc");
PlaybackDialog->Title = "Select a Playback File";
PlaybackDialog->Filter = "Remote Train Control files (*.rtc)|*.rtc";
PlaybackDialog->FileName = "*.rtc";
PlaybackDialog->Options = PlaybackDialog->Options << ofEnableSizing << ofPathMustExist << ofFileMustExist;

if (PlaybackDialog->Execute()) {
	// user pressed [Open]
	MainForm->RecordingBuffer->Clear();
	NameOfFile = PlaybackDialog->FileName;
   	// Change working directory to whatever was selected
   	MainForm->WorkingDir = ExtractFilePath(NameOfFile);
	} else {
    // user pressed [Cancel]
    return;	// user canceled the File Open Dialog
    }
if (!NameOfFile.IsEmpty()) {
	// start the playback thread - FreeOnTerminate is not set so we must delete the Thread
	ResetEvent(MainForm->TerminationPBEvent);	// reset any lingering events
    // CALLEDBYSETUP indicates that playback was started from the Setup window
			// Create suspended - in BCB 4.0, the thread constructor does not wait for lower level constructors itself.
		    // The Execute() routine actually runs before some lower level construction is finished.
			// Don't put constructions in the thread constructor. if you have to, start the thread suspended and then start it running
			// per Remy LeBeau (TeamB) it is a Known bug that was fixed in BCB 6.0
	PBThread =  new TPlaybackFileThread(true, this, Handle, RemoteNo, NameOfFile, CALLEDBYSETUP, 0);
	assert(PBThread);
	#if __BORLANDC__ >= 0x720
	PBThread->Start();			// start the thread for CB10.1
	#else
	PBThread->Resume();			// start the thread for BCB4.0 - use deprecated in CB10.1, use Start() instead
	#endif
	if (DEBUGGING) DB->Debug_Print("Starting Playback Thread for : " + NameOfFile);
    } else {
    MainForm->DebugMessage("Playback File Name is missing");
    return;
    }
MainForm->bPlayingBack = true;
// set buttons to playback state
PlaybackStartButton->Enabled = false;
PlaybackStopButton->Enabled = true;
RecordStartButton->Enabled = false;
if (DEBUGGING) DB->Debug_Print("Playback Started : " + NameOfFile);
}
//---------------------------------------------------------------------------

void __fastcall TSetup::CMStopPlayback(TMessage &Message)
{
// this message handler is called only when TPlayBackFile is called from TSetup
if (Message.LParam != CALLEDBYSETUP) return;
if (PBThread != NULL) {
	PBThread->Terminate();
	PBThread->WaitFor();	// do not use FreeOnTerminate
	delete PBThread;
	PBThread = NULL;	// release the pointer
	if (DEBUGGING) DB->Debug_Print("TSetup : CMStopPlayback : Playback Stopped");
	}

MainForm->bPlayingBack = false;
// set buttons to normal state
PlaybackStartButton->Enabled = true;
PlaybackStopButton->Enabled = false;
RecordStartButton->Enabled = true;
}
//---------------------------------------------------------------------------

void __fastcall TSetup::PlaybackStopButtonClick(TObject *Sender)
{
// turn off playback
// send a stop message to myself
(void) PostMessage(Handle, CM_STOPPLAYBACK, NULL, CALLEDBYSETUP);
}
//---------------------------------------------------------------------------

void __fastcall TSetup::DroponIdleClick(TObject *Sender)
{
if (DropDTRonIdle->Checked || DropRTSonIdle->Checked) {
	// if the port is connected then start the IdleTimer
	if (MainForm->IsTIUConnected()) {
    	MainForm->IdleTimer->Enabled = true;
        MainForm->IdleTimer->Tag = 0;	// counts 10 seconds
        }
    } else {
    MainForm->IdleTimer->Enabled = false;
    }
}
//---------------------------------------------------------------------------

void __fastcall TSetup::NumRetriesChange(TObject *Sender)
{
// NumRetries is the number of times to retry after the 1st time. Therefore, setting
// it to 2 means the command will be sent a total of 3 times.
if (NumRetries->ValueAsInt == 10) {
    MainForm->DebugMessage(" Setting '# Retries' to 10 has a special Debug meaning. Be sure you want to do this.");
	}
if (DEBUGGING) DB->Debug_Print("# Retries changed to " + String(NumRetries->ValueAsInt));
}
//---------------------------------------------------------------------------

void __fastcall TSetup::MaxSpeedChange(TObject *Sender)
{
if (DEBUGGING) DB->Debug_Print("Overall Maximum Speed changed to " + String(MaxSpeed->ValueAsInt));
}
//---------------------------------------------------------------------------

void __fastcall TSetup::ModeSelectClick(TObject *Sender)
{
if (ModeSelect->ItemIndex == 0) {
	// Wired Mode
    COMSelGroup->Caption = "Serial Port for TIU" + String(TIUNo);
    COMLabel->Caption = "Wired COM";
    DropDTRonIdle->Visible = true;
    DropRTSonIdle->Visible = true;
    MainForm->Caption = "Remote Train Control (Wired)";
    return;
    }
if (ModeSelect->ItemIndex == 1) {
	// Radio Mode
    COMSelGroup->Caption = "Serial Port for Radio";
    COMLabel->Caption = "Radio COM";
    DropDTRonIdle->Visible = false;
    DropRTSonIdle->Visible = false;
    MainForm->Caption = "Remote Train Control (Radio/WiFi)";
    return;
    }
if (ModeSelect->ItemIndex == 2) {
	// TIU EMulator Mode
    COMSelGroup->Caption = "Serial Port for TIU Emulator";
    COMLabel->Caption = "Emulator COM";
    DropDTRonIdle->Visible = false;
    DropRTSonIdle->Visible = false;
    MainForm->Caption = "Remote Train Control (Emulator)";
    return;
    }
}
//---------------------------------------------------------------------------

#include <setupapi.h>

#define MAX_PORT_NUM      (256)
#define MAX_STR_LEN       (256 * sizeof(TCHAR))

// see http://blog.csdn.net/u013606170/article/details/46903713
// and
// http://www.naughter.com/enumser.html
/*
Copyright (c) 1998 - 2013 by PJ Naughter (Web: www.naughter.com, Email: pjna@naughter.com)

All rights reserved.

Copyright / Usage Details:

You are allowed to include the source code in any product (commercial, shareware, freeware or otherwise)
when your product is released in binary form. You are allowed to modify the source code in any way you want
except you cannot modify the copyright details at the top of each module. If you want to distribute source
code with your application, then you are only allowed to distribute versions released by the author. This is
to maintain a single distribution point for the source code.

*/

BOOL __fastcall TSetup::EnumerateComPorts(void) { 
 int ret, Number;

 TCHAR Mess[MAX_STR_LEN];
 GUID *pGuid;
 DWORD dwGuids;
 HDEVINFO hDevInfoSet;
 TCHAR portString[10];

 COMCheckListBox->Clear();

 BOOL bMoreItems;
 SP_DEVINFO_DATA devInfo;

 ret = FALSE;

 //First need to convert the name "Ports" to a GUID using SetupDiClassGuidsFromName
 dwGuids = 0;
 SetupDiClassGuidsFromName(TEXT("Ports"), NULL, 0, &dwGuids);

 if(0 == dwGuids) return FALSE;
 
 //Allocate the needed memory
 pGuid = (GUID*)HeapAlloc(GetProcessHeap(), HEAP_GENERATE_EXCEPTIONS, dwGuids * sizeof(GUID));
 if(NULL == pGuid) return FALSE;

 //Call the function again
 if (FALSE == SetupDiClassGuidsFromName(TEXT("Ports"), pGuid, dwGuids, &dwGuids)) {
  return FALSE;
  }/*if*/

 hDevInfoSet = SetupDiGetClassDevs(pGuid, NULL, NULL, DIGCF_PRESENT /*| DIGCF_DEVICEINTERFACE*/);
 if (INVALID_HANDLE_VALUE == hDevInfoSet) {
  //Set the error to report
  _sntprintf(Mess, sizeof Mess, TEXT("error SetupDiGetClassDevs, %d"), GetLastError());
  if (DEBUGGING) DB->Debug_Print(Mess);
  return FALSE;
  }/*if */

 devInfo.cbSize = sizeof(SP_DEVINFO_DATA);
 unsigned int i = 0;
 unsigned int jj = 0;

 do {
  HKEY hDeviceKey;
  BOOL isFound;

  isFound = FALSE;
  bMoreItems = SetupDiEnumDeviceInfo(hDevInfoSet, i, &devInfo);
  if(FALSE == bMoreItems) break;

  i++;

  hDeviceKey = SetupDiOpenDevRegKey(hDevInfoSet, &devInfo, DICS_FLAG_GLOBAL, 0, DIREG_DEV, KEY_QUERY_VALUE);

  if (INVALID_HANDLE_VALUE != hDeviceKey) {
   size_t nLen;
   LPTSTR pszPortName;

   {
    //First query for the size of the registry value
    DWORD dwType;
    DWORD dwDataSize;
    LONG err;
    DWORD dwAllocatedSize;
    DWORD dwReturnedSize;
    dwType = 0; dwDataSize = 0;

    err = RegQueryValueEx(hDeviceKey, TEXT("PortName"), NULL, &dwType, NULL, &dwDataSize);
    if (ERROR_SUCCESS != err) continue;
   
    //Ensure the value is a string
    if (dwType != REG_SZ) continue;
   
    //Allocate enough bytes for the return value
	dwAllocatedSize = dwDataSize + sizeof(TCHAR);
	// "+ sizeof(TCHAR)" is to allow us to NULL terminate
    // the data if it is not null terminated in the registry

    pszPortName = (LPTSTR)LocalAlloc(LMEM_FIXED, dwAllocatedSize);
    if (pszPortName == NULL) continue;

    //Recall RegQueryValueEx to return the data
    pszPortName[0] = '\0';
    dwReturnedSize = dwAllocatedSize;
    err = RegQueryValueEx(hDeviceKey, TEXT("PortName"), NULL, &dwType, (LPBYTE)pszPortName, &dwReturnedSize);
    if (ERROR_SUCCESS != err) {
     LocalFree(pszPortName);
     continue;
     }

    //Handle the case where the data just returned is the same size as the allocated size. This could occur where the data
    //has been updated in the registry with a non null terminator between the two calls to ReqQueryValueEx above. Rather than
    //return a potentially non-null terminated block of data, just fail the method call
    if (dwReturnedSize >= dwAllocatedSize) continue;

    //NULL terminate the data if it was not returned NULL terminated because it is not stored null terminated in the registry
    if (pszPortName[dwReturnedSize/sizeof(TCHAR) - 1] != '\0') pszPortName[dwReturnedSize/sizeof(TCHAR)] = '\0';
   }/*local varable*/

   //If it looks like "COMX" then
   //add it to the array which will be returned
   nLen = _tcsclen(pszPortName);

   if (3 < nLen){
   	if (0 == _tcsnicmp(pszPortName, TEXT("COM"), 3)) {
      if(FALSE == _istdigit(pszPortName[3]) ) continue;
      //Work out the port number
      _sntprintf(Mess, sizeof Mess, TEXT("Port %s"), pszPortName);
 	  if (DEBUGGINGX(7)) DB->Debug_Print(Mess);
      //COMCheckListBox->Items->Add(pszPortName);
      _tcsnccpy(portString, pszPortName, sizeof(portString));
      } else {
      continue;
      }/*if 0 == _tcsnicmp(pszPortName, TEXT("COM"), 3)*/
	}/*if 3 < nLen*/

   LocalFree(pszPortName);    
   isFound = TRUE;  

   //Close the key now that we are finished with it
   RegCloseKey(hDeviceKey);
  }/*INVALID_HANDLE_VALUE != hDeviceKey*/

  if(FALSE == isFound)
   continue;

  //If the port was a serial port, then also try to get its friendly name  
  {
   TCHAR szFriendlyName[1024];     
   DWORD dwSize;
   DWORD dwType;
   szFriendlyName[0] = '\0';
   dwSize = sizeof(szFriendlyName);
   dwType = 0;

   if((TRUE == SetupDiGetDeviceRegistryProperty(hDevInfoSet, &devInfo, SPDRP_DEVICEDESC, &dwType, (PBYTE)(szFriendlyName), dwSize, &dwSize) ) && (REG_SZ == dwType)) {
    _sntprintf(Mess, sizeof Mess, TEXT("%s %s"), portString, szFriendlyName);
 	if (DEBUGGINGX(7)) DB->Debug_Print(Mess);
    COMCheckListBox->Items->Add(Mess);
    } else {
 	_sntprintf(Mess, sizeof Mess, TEXT("%s"), portString);
 	if (DEBUGGINGX(7)) DB->Debug_Print(Mess);
    COMCheckListBox->Items->Add(Mess);
    }/*if SetupDiGetDeviceRegistryProperty */
  }/*local variable */

  jj++;
  } while(1);

 HeapFree(GetProcessHeap(), 0, pGuid);
 Number = jj;
 _sntprintf(Mess, sizeof Mess, TEXT("Found %d Ports"), Number);
 if (DEBUGGINGX(7)) DB->Debug_Print(Mess);
 //
 // if one of the ports is the port that is already selected, then highlight it.
 //
 COMCheckListBox->ItemIndex = -1;
 for (int cc = 0 ; cc < Number ; cc++) {
    int nPort;
	_stscanf(COMCheckListBox->Items->Strings[cc].c_str(),TEXT("COM%d"), &nPort);
    if (nPort == PortNum->ValueAsInt) {
        COMCheckListBox->ItemIndex = cc;
    	break;
    	}
 	}

 if(0 < jj) ret = TRUE;
 return ret;
}
//---------------------------------------------------------------------------

void __fastcall TSetup::EnumerateSerialPorts1Click(TObject *Sender)
{
(void) EnumerateComPorts();
}
//---------------------------------------------------------------------------

void __fastcall TSetup::COMCheckListBoxClick(TObject *Sender)
{
const int SelectedCOMPort = COMCheckListBox->ItemIndex;
int nPort;
_stscanf(COMCheckListBox->Items->Strings[SelectedCOMPort].c_str(),TEXT("COM%d"), &nPort);
PortNum->Value = nPort;
}
//---------------------------------------------------------------------------

void __fastcall TSetup::PortNumValueChanged(TObject *Sender)
{
//
// if one of the ports is the port that is selected, then highlight it.
//
if (COMCheckListBox->Items->Count != 0) {	// for non empty list
	COMCheckListBox->ItemIndex = -1;
	for (int cc = 0 ; cc < COMCheckListBox->Items->Count ; cc++) {
    	int nPort;
		_stscanf (COMCheckListBox->Items->Strings[cc].c_str(),TEXT("COM%d"), &nPort);
    	if (nPort == PortNum->ValueAsInt) {
    	    COMCheckListBox->ItemIndex = cc;
    		break;
    		}
 		}
    }

// Update labels on MainForm
char PortName[10];
snprintf(PortName, sizeof PortName, "COM%d", PortNum->ValueAsInt);
MainForm->ConnectedRadioLED->Caption = PortName;
//Radio_COM_Mode->Checked = true;
//RFID_WiFi_Mode->Checked = false;
MainForm->TIUBufferLabel->Caption = "TIU Buffer " + MainForm->ConnectedRadioLED->Caption;

}
//---------------------------------------------------------------------------

void __fastcall TSetup::RFIDNumValueChanged(TObject *Sender)
{
//
//RFID_COM_Mode->Checked = true;
//Radio_COM_Mode->Checked = false;

// Update labels on MainForm
char RFIDName2[10];
snprintf(RFIDName2, sizeof RFIDName2, "COM%d", RFIDNum->ValueAsInt);
MainForm->ConnectedRFIDLED->Caption = RFIDName2;
MainForm->RFIDBufferLabel->Caption = "RFID Buffer " + MainForm->ConnectedRFIDLED->Caption;
MainForm->RFIDBuffer->Enabled = true;
MainForm->RFIDBuffer->Color = clWindow;
}
//---------------------------------------------------------------------------

void __fastcall TSetup::FormShow(TObject *Sender)
{
// Enumerate COM ports
(void) EnumerateComPorts();
// in debugging mode, make a few more controls visible
if (DEBUGGING) {
	Radio_WiFi_Multicast->Visible = true;
	AIU_WiFi_Multicast->Visible = true;
    } else {
	Radio_WiFi_Multicast->Visible = false;
	AIU_WiFi_Multicast->Visible = false;
    }
}
//---------------------------------------------------------------------------

void __fastcall TSetup::WMDropFiles(TWMDropFiles &message)
{
if (!MainForm->IsTIUConnected()) {
	MainForm->DebugMessagePopup("Warning", MainForm->ConnectMsg, "");
	MainForm->bPlayingBack = false;
	return;
    }
if (MainForm->bRecording) {
    MainForm->DebugMessage("You must stop Recording before you can Playback a previous Recording");
	return;
    }

// Process first valid filename found
UINT filecount = DragQueryFile((HDROP) message.Drop, 0xFFFFFFFF, NULL, 0);
if ((PBThread == NULL) && (filecount > 0)) {
	String FullFileName;
	for (UINT ii = 0 ; ii < filecount ; ++ii){
    	Application->ProcessMessages();		// ok to use this during DragQueryFile() processing
		// Get the file name
		char FileName[MAX_PATH];
		DragQueryFileA ((HDROP) message.Drop, ii, FileName, MAX_PATH);
		FullFileName = FileName;
		//Application->MessageBox(FullFileName.c_str(), TEXT("Result of DragQueryFile()"), MB_OKCANCEL | MB_DEFBUTTON1);
		if (FileExists(FullFileName)) {
        	String Extension = LowerCase(ExtractFileExt(FullFileName));
            if (Extension == ".rtc") {
				// FIXME do something for each file
    	        //DebugBuffer->Lines->Append("Dropped file " + FullFileName);
                // Change working directory to whatever was selected
				MainForm->WorkingDir = ExtractFilePath(FullFileName);
				MainForm->RecordingBuffer->Lines->LoadFromFile(FullFileName);
				if (DEBUGGING) DB->Debug_Print("ReadFromDisk (Playback): " + FullFileName);
				if (!FullFileName.IsEmpty()) {
					// start the playback thread - FreeOnTerminate is not set so we must delete the Thread
				    // worry about deleting the Thread
					ResetEvent(MainForm->TerminationPBEvent);	// reset any lingering events
					// Create suspended - in BCB 4.0, the thread constructor does not wait for lower level constructors itself.
				    // The Execute() routine actually runs before some lower level construction is finished.
				    // Don't put constructions in the thread constructor. if you have to, start the thread suspended and then start it running
				    // per Remy LeBeau (TeamB) it is a Known bug that was fixed in BCB 6.0
					PBThread = new TPlaybackFileThread(true, this, Handle, RemoteNo, FullFileName, CALLEDBYSETUP, 0);
					assert(PBThread);
					#if __BORLANDC__ >= 0x720
					PBThread->Start();			// start the thread for CB10.1
					#else
					PBThread->Resume();			// start the thread for BCB4.0 - use deprecated in CB10.1, use Start() instead
					#endif
                    if (DEBUGGING) DB->Debug_Print("Starting Playback Thread for : " + FullFileName);

                    } else {
                    MainForm->DebugMessage("Playback File Name is missing");
                    }
				MainForm->bPlayingBack = true;
				// set buttons to playback state
				PlaybackStartButton->Enabled = false;
				PlaybackStopButton->Enabled = true;
				RecordStartButton->Enabled = false;
				if (DEBUGGING) DB->Debug_Print("Playback Started : " + FullFileName);
                break;
				}
            if (DEBUGGING) DB->Debug_Print("Not a Recording File : " + FullFileName);
			continue;
			} else {
			continue;
			}
		}
	//String FileName = LowerCase(ExtractFileName(FullFileName));
	// use last dropped filename
	//FileEdit->Text = FullFileName;
	}
DragFinish ((HDROP) message.Drop);
return;
}
//---------------------------------------------------------------------------

void __fastcall TSetup::ManuallysetSpeed1Click(TObject *Sender)
{
// enable the baud rate selection radio buttons
SerialPortListGroup->Visible = false;
BaudRateSel->Visible = true;
BitsSel->Visible = true;
ParitySel->Visible = true;
StopSel->Visible = true;
BaudRateSel->Enabled = true;
BitsSel->Enabled = true;
ParitySel->Enabled = true;
StopSel->Enabled = true;
}
//---------------------------------------------------------------------------

void __fastcall TSetup::AutomaticallysetSpeed1Click(TObject *Sender)
{
// disable the baud rate selection radio buttons and return rates to default values
SerialPortListGroup->Visible = true;
BaudRateSel->ItemIndex = 2;
BitsSel->ItemIndex = 1;
ParitySel->ItemIndex = 0;
StopSel->ItemIndex = 0;
BaudRateSel->Visible = false;
BitsSel->Visible = false;
ParitySel->Visible = false;
StopSel->Visible = false;
BaudRateSel->Enabled = false;
BitsSel->Enabled = false;
ParitySel->Enabled = false;
StopSel->Enabled = false;
}
//---------------------------------------------------------------------------

void __fastcall TSetup::FormKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
// Make sure KeyPreview is true for this form
assert(KeyPreview);
// Pass function keys to the MainForm, handle rest of the keys here.
switch (Key) {
    case VK_F1:
    case VK_F2:
    case VK_F3:
    case VK_F4:
    case VK_F5:
    case VK_F6:
    case VK_F7:
    case VK_F8:
    case VK_F9:
    case VK_F10:
    case VK_F11:
    case VK_F12:
		MainForm->FormKeyDown(Sender, Key, Shift);
        Key = 0;	// handled
		break;
	}	
}
//---------------------------------------------------------------------------

void __fastcall TSetup::DisableEmergencyStopButtonClick(TObject *Sender)
{
ShowMessage("Restart required for new setting to take effect");
}
//---------------------------------------------------------------------------

void __fastcall TSetup::SuperModeSetupButtonClick(TObject *Sender)
{
// create the Super TIU window
// This code should work with auto create on OR off.
// ------------ With programmed create using C++ Builder's generated pointer
// Super TIU Form should be manually created
if (!SuperTIU) {	// since I used the C++ pointer, it will get automatically deleted at close
	SuperTIU = new TSuperTIU(this);     // Pass "this" as the owner
	assert(SuperTIU != NULL);			// make sure that construction worked
	}
if (SuperTIU) SuperTIU->Show();
}
//---------------------------------------------------------------------------

void __fastcall TSetup::TransparentSpinChange(TObject *Sender)
{
if (DEBUGGING) DB->Debug_Print("Tranparency changed to " + String(TransparentSpin->ValueAsInt));
}
//---------------------------------------------------------------------------

void __fastcall TSetup::Ignore_FFClick(TObject *Sender)
{
Ignore_FF_Char = Ignore_FF->Checked;
}
//---------------------------------------------------------------------------

void __fastcall TSetup::EncodedRecordFormatClick(TObject *Sender)
{
MainForm->EncodedRecordFormat = EncodedRecordFormat->Checked;	// put format switch into global variable
}
//---------------------------------------------------------------------------

void __fastcall TSetup::PlaybackDialogShow(TObject *Sender)
{
// change Open button to "Playback"
// http://www.bcbjournal.org/articles/vol3/9912/Changing_common_dialog_button_captions.htm
HWND hWndDlg = GetParent(PlaybackDialog->Handle);
if (hWndDlg == NULL) {
	if (DEBUGGING) DB->Debug_Print("GetParent() returned NULL");
    LPVOID lpMsgBuf;
    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
        NULL,
        GetLastError(),
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
        (LPTSTR) &lpMsgBuf,
        0,
        NULL);

    // Display the string.
	if (DEBUGGING) DB->Debug_Print(String((char *)lpMsgBuf));
    // Free the buffer.
    LocalFree( lpMsgBuf );
	} else {
	HWND hWndBtn = GetDlgItem(hWndDlg, IDOK); //1);
	if (hWndBtn == NULL) {
		if (DEBUGGING) DB->Debug_Print("Control does not exist");
		} else {
		SetWindowText(hWndBtn, TEXT("Playback"));
    	}
#ifdef CANCELBUT
	hWndBtn = GetDlgItem(hWndDlg, IDCANCEL); //2);
	if (hWndBtn == NULL) {
		if (DEBUGGING) DB->Debug_Print("Control does not exist");
		} else {
		SetWindowText(hWndBtn, TEXT("Cancel"));
    	}
#endif
    }
}
//---------------------------------------------------------------------------
#ifdef USEFULINFO
#define IDOK                1
#define IDCANCEL            2
#define IDABORT             3
#define IDRETRY             4
#define IDIGNORE            5
#define IDYES               6
#define IDNO                7
#if(WINVER >= 0x0400)
#define IDCLOSE         8
#define IDHELP          9
#endif /* WINVER >= 0x0400 */
#endif
//---------------------------------------------------------------------------

void __fastcall TSetup::MPHScalingChange(TObject *Sender)
{
MainForm->GlobalMPHScaling = MPHScaling->Value;	
}
//---------------------------------------------------------------------------

void __fastcall TSetup::CopyrightClick(TObject *Sender)
{
// Copy the Form to the clipboard.
// In BCB4.0, there is no SetSize method associated with TBitmap
// so we have to use Width and Height separately.
//
// see:
// https://stackoverflow.com/questions/3190777/how-do-i-copy-a-form-as-an-image-to-the-clipboard
//
Graphics::TBitmap * bm;
bm = new Graphics::TBitmap;
try
	{
    #if __BORLANDC__ >= 0x720
	// Defined for CB10.1 C++ programs that use the VCL.
	bm->SetSize(ClientWidth, ClientHeight);
    #else
    bm->Width = ClientWidth;
    bm->Height = ClientHeight;
    #endif
	BitBlt(bm->Canvas->Handle, 0, 0, ClientWidth, ClientHeight, Canvas->Handle, 0, 0, SRCCOPY);
    Clipboard()->Assign(bm);
    }
__finally
	{
	delete bm;
    }
}
//---------------------------------------------------------------------------

void __fastcall TSetup::DTRAlwaysOnClick(TObject *Sender)
{
MainForm->DTRSwitch->Checked = DTRAlwaysOn->Checked;
}
//---------------------------------------------------------------------------

void __fastcall TSetup::Scan(String Service, TListBox * MyListBox)
{
// Scan for ESP8266 devies with a given Service
// Set USEmDNSThread to true to use mDNSThread, set to false to use mDNSSServices
// mDNSServices get some kind of memory corruption problem in open_client_sockets()
// during the call to GetAdaptersAddresses when using the new code and a failure
// in the OLD code around the call to ipv4_address_to_string()
#define USEmDNSThread true
//
#if USEmDNSThread

// Create (but don't run) the MultiDNS Thread
// Create suspended - in BCB 4.0, the thread constructor does not wait for lower level constructors itself.
// The Execute() routine actually runs before some lower level construction is finished.
// Don't put constructions in the thread constructor. if you have to, start the thread suspended and then start it running
// per Remy LeBeau (TeamB) it is a Known bug that was fixed in BCB 6.0

uint32_t S_Addr = WSocket_inet_addr(Setup->IPComboBox->Items->Strings[Setup->IPComboBox->ItemIndex].c_str());	// Previously set to INADDR_ANY
ptrMultiDNSThread = new TMultiDNS(true, this, Handle, S_Addr, MyListBox->Items);
assert(ptrMultiDNSThread);	// Check that constructor worked

// Set variables used by the mDNS thread
ptrMultiDNSThread->query_arg =  Service;

// Successful Open
#if __BORLANDC__ >= 0x720
ptrMultiDNSThread->Start();		// start the thread for CB10.1
#else
ptrMultiDNSThread->Resume();	// start the thread for BCB4.0 - use deprecated in CB10.1, use Start() instead
#endif
ptrMultiDNSThread->WaitFor();
delete ptrMultiDNSThread;
ptrMultiDNSThread = NULL;
#else   // else use mDNSServices
if (!ptrmDNSServices) {
    //-------------------- Start mDNS Server ---------------------------
    if (Setup->Radio_WiFi_Mode->Checked || Setup->RFID_WiFi_Mode->Checked || Setup->AIU_WiFi_Mode->Checked) {
        // Start the mDNS responder - mDNS query if in WiFi mode
       	Setup->StartmDNSServices(Setup->RFID_WiFi_Mode->Checked, Setup->Radio_WiFi_Mode->Checked);
        }
    }
ptrmDNSServices->mDNSQuery(Service);

#endif

}
//---------------------------------------------------------------------------

void __fastcall TSetup::WiFiListBoxClick(TObject *Sender)
{
TCHAR IP[50], Name[50];
int Port;
if (WiFiListBox->Items->Count == 0) {
	return;
    }
const int SelectedIP = WiFiListBox->ItemIndex;

int retv = _stscanf(WiFiListBox->Items->Strings[SelectedIP].c_str(), TEXT("%s %d %s"), IP, &Port, Name);
//
// Note that I have found that Multicast mode gives much slower response times
// than Unicast. Multicast is not needed with the Radio since only one
// radio is supported. The check box is not visible and cannot be changed.
//
// This is different than the AIU devices. There can be an almost unlimited number of
// AIU devices so that Multicast is required. The check box is not visible and cannot be
// changed.
//
if (retv == 3) {
	RadioIP->Text = String(IP);
	RadioPort->Text = Port;
	if (DEBUGGING) DB->Debug_Print("WiFi Selected IP : " + RadioIP->Text + " : Port : " + RadioPort->Text);
    } else {
    RadioIP->Text = "239.255.200.3";
	RadioPort->Text = "8890";
    if (DEBUGGING) DB->Debug_Print("No RTC Modem over WiFi Found, using Multicast IP");
    }

}
//---------------------------------------------------------------------------

void __fastcall TSetup::AIUWiFiListBoxClick(TObject *Sender)
{
//
// Note that I have found that Multicast mode gives much slower response times
// than Unicast. Multicast is not needed with the Radio since only one
// radio is supported. The check box is not visible and cannot be changed.
//
// This is different than the AIU devices. There can be an almost unlimited number of
// AIU devices so that Multicast is required. The check box is not visible and cannot be
// changed.
//
    AIUIP->Text = "239.255.200.3";
	AIUPort->Text = "8890";
}
//---------------------------------------------------------------------------

void __fastcall TSetup::ScanRadioWiFiButtonClick(TObject *Sender)
{
if (!Radio_WiFi_Mode->Checked) {
	MainForm->DebugMessagePopup("Warning", "Radio Modem WiFi must be enabled before performing this operation");
    return;
	}
WiFiListBox->Clear();
ScanModemClick(Sender);
}
//---------------------------------------------------------------------------

void __fastcall TSetup::ScanAIUWiFiButtonClick(TObject *Sender)
{
if (!AIU_WiFi_Mode->Checked) {
	MainForm->DebugMessagePopup("Warning", "AIU WiFi must be enabled before performing this operation");
    return;
	}
AIUWiFiListBox->Clear();
// clear boolean flags
for (int iTIU = 1; iTIU <= TIUMAX ; iTIU++) { // TIU 1-5
    AIUEmulator[iTIU-1].Found = false;
	}
for (int cnt = 1; cnt <= 100 ; cnt++) { // Display 1 - 100
    MessageDisplay[cnt-1].Found = false;
	}
// Scan for AIU capable devices
MainForm->DebugMessagePopup("Wait", "Scanning for RTCAIU WiFi Devices");
Scan("_RTCAIU._udp.local", AIUWiFiListBox);
// Scan for Message Display Devices
MainForm->DebugMessagePopup("Wait", "Scanning for RTCMSG WiFi Devices");
Scan("_RTCMSG._udp.local", AIUWiFiListBox);

// parse the results of the mDNS scan to find all TIU that have AIU Emulators
TCHAR IP[50], Name[50];
int Port;
// Set variables used by the mDNS thread
if (AIUWiFiListBox->Items->Count > 0) {
    for (int idex = 0 ; idex < AIUWiFiListBox->Items->Count ; idex++) {
        int retv = _stscanf(AIUWiFiListBox->Items->Strings[idex].c_str(), TEXT("%s %d %s"), IP, &Port, Name);
        if (retv == 3) {	// found three items
            // parse Name   AIU411 : 4 = TIU  1 = AIU  1 = Starting Channel
            int TIU, AIU, Chan;
            retv = _stscanf(Name, TEXT("AIU-%1d%1d%d"), &TIU, &AIU, &Chan);	// parse the Name to get TIU number
            if (retv == 3) {	// found three items
                // set AIUEmulator[] indexed by TIU-1
                AIUEmulator[TIU-1].Found = true;	// Update the AIUEmulator boolean array flag for each TIU found
                _tcsncpy(AIUEmulator[TIU-1].IP, IP, sizeof IP);
                AIUEmulator[TIU-1].Port = Port;
                _tcsncpy(AIUEmulator[TIU-1].Name, Name, sizeof Name);
				AIUEmulator[TIU-1].AIU = AIU;
                AIUEmulator[TIU-1].Chan = Chan;
                continue;
                }
            int Display;
            retv = _stscanf(Name, TEXT("Display-%d"), &Display);
            if (retv == 1)  {    // found one item
                // set MessageDisplay[] indexed by Display-1
                MessageDisplay[Display-1].Found = true;
                _tcsncpy(MessageDisplay[Display-1].IP, IP, sizeof IP);
                MessageDisplay[Display-1].Port = Port;
                _tcsncpy(MessageDisplay[Display-1].Name, Name, sizeof Name);
                continue;
                }
            }
        }
    }
}
//---------------------------------------------------------------------------

void __fastcall TSetup::ScanSerialButtonClick(TObject *Sender)
{
// Find all serial ports
MainForm->DebugMessagePopup("Wait", "Scanning for USB Serial Ports");
(void) EnumerateComPorts();
}
//---------------------------------------------------------------------------

void __fastcall TSetup::ScanModemClick(TObject *Sender)
{
// Scan for Modem capable devices
MainForm->DebugMessagePopup("Wait", "Scanning for RTCModem WiFi Devices");
// Set variables used by the mDNS thread
Scan("_RTCModem._udp.local", WiFiListBox);
if (WiFiListBox->Items->Count == 0) {
	MainForm->DebugMessagePopup("Error", "No RTCModem WiFi Devices found. RTC requires exactly one RTCModem Device");
	} else
    if (WiFiListBox->Items->Count > 1) {
		MainForm->DebugMessagePopup("Warning", "Too Many RTCModem WiFi Devices. Click on one of the listed RTCModem Devices to use it");
    	} else
    	MainForm->DebugMessagePopup("Information", "Click on the listed RTCModem WiFi Device to use it");
}
//---------------------------------------------------------------------------

void __fastcall TSetup::ScanOTAClick(TObject *Sender)
{
WiFiListBox->Clear();
// Scan for OTA capable devices
MainForm->DebugMessagePopup("Wait", "Scanning for OTA Capable Devices");
// Set variables used by the mDNS thread
Scan("_arduino._tcp.local", WiFiListBox);
}
//---------------------------------------------------------------------------

void __fastcall TSetup::ScanMSGClick(TObject *Sender)
{
WiFiListBox->Clear();
// Scan for RTCMSG capable devices
MainForm->DebugMessagePopup("Wait", "Scanning for RTCMSG WiFi Devices");
// Set variables used by the mDNS thread
Scan("_RTCMSG._udp.local", WiFiListBox);
}
//---------------------------------------------------------------------------

void __fastcall TSetup::ScanRFIDudpClick(TObject *Sender)
{
WiFiListBox->Clear();
// Scan for RFID udp Detector capable devices
MainForm->DebugMessagePopup("Wait", "Scanning for RTCNFC udp WiFi Devices");
// Set variables used by the mDNS thread
Scan("_RTCRFID._udp.local", WiFiListBox);
}
//---------------------------------------------------------------------------

void __fastcall TSetup::ScanRFIDtcpClick(TObject *Sender)
{
WiFiListBox->Clear();
// Scan for RFID tcp Detector capable devices
MainForm->DebugMessagePopup("Wait", "Scanning for RTCNFC tcp WiFi Devices");
// Set variables used by the mDNS thread
Scan("_RTCRFID._tcp.local", WiFiListBox);
}

//---------------------------------------------------------------------------

void __fastcall TSetup::ScanMasterClick(TObject *Sender)
{
WiFiListBox->Clear();
// Scan for RTC Master capable devices
MainForm->DebugMessagePopup("Wait", "Scanning for RTCMaster");
// Set variables used by the mDNS thread
Scan("_RTCMaster._udp.local", WiFiListBox);
}
//---------------------------------------------------------------------------

void __fastcall TSetup::ScanHostClick(TObject *Sender)
{
WiFiListBox->Clear();
// Scan for RTC Master capable devices
MainForm->DebugMessagePopup("Wait", "Scanning for RTCHost");
// Set variables used by the mDNS thread
Scan("_RTCHost._udp.local", WiFiListBox);
}
//---------------------------------------------------------------------------

void __fastcall TSetup::ScanAIUClick(TObject *Sender)
{
WiFiListBox->Clear();
// Scan for AIU capable devices
MainForm->DebugMessagePopup("Wait", "Scanning for RTCAIUWiFi Devices");
// Set variables used by the mDNS thread
Scan("_RTCAIU._udp.local", WiFiListBox);}
//---------------------------------------------------------------------------

void __fastcall TSetup::ScanEngineClick(TObject *Sender)
{
WiFiListBox->Clear();
// Scan for RTCEngine capable devices
MainForm->DebugMessagePopup("Wait", "Scanning for RTCEngine WiFi Devices");
// Set variables used by the mDNS thread
Scan("_RTCEngine._udp.local", WiFiListBox);
}
//---------------------------------------------------------------------------

void __fastcall TSetup::OOKRadioHelpClick(TObject *Sender)
{
ShellExecute(NULL, TEXT("open"), TEXT("http://www.silogic.com/trains/OOK_Radio_Support.html"), NULL, NULL, SW_SHOWNORMAL);
}
//---------------------------------------------------------------------------

void __fastcall TSetup::RFIDTagsHelp1Click(TObject *Sender)
{
ShellExecute(NULL, TEXT("open"), TEXT("http://www.silogic.com/trains/RFID.html"), NULL, NULL, SW_SHOWNORMAL);
}
//---------------------------------------------------------------------------

void __fastcall TSetup::RTCHelp1Click(TObject *Sender)
{
ShellExecute(NULL, TEXT("open"), TEXT("http://www.silogic.com/trains/RTC_Running.html"), NULL, NULL, SW_SHOWNORMAL);
}
//---------------------------------------------------------------------------

void __fastcall TSetup::WiFiInformation1Click(TObject *Sender)
{
ShellExecute(NULL, TEXT("open"), TEXT("http://www.silogic.com/trains/WiFi.html"), NULL, NULL, SW_SHOWNORMAL);
}
//---------------------------------------------------------------------------

void __fastcall TSetup::StartmDNSServices(bool RTCMaster, bool RTCHost)
{
// Start advertising the Services RTCHost and RTCMaster - both services are started, parameters not used
//
// Note: this routine always advertises the RTCMaster and the RTCHost services. Hard coded.
if (ptrmDNSServices) {	// already running
	StopmDNSServices(RTCMaster, RTCHost);	// stop it
	}
// Create (but don't run) the MultiDNS Thread
// Create suspended - in BCB 4.0, the thread constructor does not wait for lower level constructors itself.
// The Execute() routine actually runs before some lower level construction is finished.
// Don't put constructions in the thread constructor. if you have to, start the thread suspended and then start it running
// per Remy LeBeau (TeamB) it is a Known bug that was fixed in BCB 6.0
// Select a network
if (IPComboBox->ItemIndex == -1 || IPComboBox->Items->Count == 0) {
    if (DEBUGGING) DB->Debug_Print("Select a Network for WiFi Operations");
	// No IP address selected or no IP addresses exist
    ScanHostIPButtonClick(NULL);
    return;
    }

uint32_t S_Addr = WSocket_inet_addr(IPComboBox->Text.c_str());	// Previously set to INADDR_ANY
mDNSTerminate = false;
if (DEBUGGING) DB->Debug_Print("Starting mDNS Service at Address = " + IPComboBox->Text);
ptrmDNSServices = new mDNSServices(true, S_Addr, RTCMaster, RTCHost);
assert(ptrmDNSServices);	// Check that constructor worked
MainForm->mDNSLED->Checked = true;
//
// Successful Open
if (ptrmDNSServices) {
	#if __BORLANDC__ >= 0x720
	ptrmDNSServices->Start();	// start the thread for CB10.1
	#else
	ptrmDNSServices->Resume();	// start the thread for BCB4.0 - use deprecated in CB10.1, use Start() instead
	#endif
    }
if (DEBUGGING) DB->Debug_Print("mDNS Server Started");
mDNSServer->Checked = true;
}
//---------------------------------------------------------------------------
void __fastcall TSetup::StopmDNSServices(bool RTCMaster, bool RTCHost)
{
// Stop advertising the Services - stops all services, parameters not used
//
if (!ptrmDNSServices) return;	// already stopped
mDNSTerminate = true;			// mDNSServices thread will free on terminate
ptrmDNSServices = NULL;
MainForm->mDNSLED->Checked = false;
if (DEBUGGING) DB->Debug_Print("mDNS Server Stopped");
mDNSServer->Checked = false;
}
//---------------------------------------------------------------------------

void __fastcall TSetup::mDNSServerClick(TObject *Sender)
{
// This is supplied as a MenuItem selection for Debugging Only.
if (!ptrmDNSServices) {
	// Start the mDNS Server
    MainForm->DebugMessagePopup("Please Wait", "Starting the mDNS Server", "");
	StartmDNSServices(RFID_WiFi_Mode->Checked, Radio_WiFi_Mode->Checked);
    } else {
	// Stop the mDNS Server
	MainForm->DebugMessagePopup("Please Wait", "Stopping the mDNS Server", "");
	StopmDNSServices(RFID_WiFi_Mode->Checked, Radio_WiFi_Mode->Checked);
    }
}
//---------------------------------------------------------------------------

void __fastcall TSetup::Radio_WiFi_MulticastClick(TObject *Sender)
{
// Radio_WiFi_MultiCast->Checked is saved in the Registry
if (Radio_WiFi_Multicast->Checked) {	// Multicast IP and Port
	RadioIP->Text = "239.255.200.3";
	RadioPort->Text = "8890";
    } else {							// Unicast IP and Port
	RadioIP->Text = "0.0.0.0";
	RadioPort->Text = "8890";
    }
}
//---------------------------------------------------------------------------

void __fastcall TSetup::AIU_WiFi_ModeClick(TObject *Sender)
{
//
// Setup AIU Emulator WiFi Mode
//
if (AIU_WiFi_Mode->Checked) {
    AIUWiFiLabel->Enabled = true;
    AIU_WiFi_Multicast->Enabled = true;
    AIUIP->Enabled = true;
    AIUPort->Enabled = true;
    AIUWiFiListBox->Enabled = true;
   	//StartmDNSServices(RFID_WiFi_Mode->Checked, Radio_WiFi_Mode->Checked);	// Start the mDNS responder if in WiFi mode
    HostIPPage->Enabled = true;
	} else {
    AIUWiFiLabel->Enabled = false;
    AIU_WiFi_Multicast->Enabled = false;
    AIUIP->Enabled = false;
    AIUPort->Enabled = false;
    AIUWiFiListBox->Enabled = false;
    }
Set_Comment();
}
//---------------------------------------------------------------------------
int __fastcall TSetup::FindIPs()
{
// https://tangentsoft.net/wskfaq/examples/ipaddr.html

char ac[80];
int IPCount = 0;

WSAData wsaData;
if (WSAStartup(MAKEWORD(1, 1), &wsaData) != 0) {
    return -1;
    }

if (gethostname(ac, sizeof(ac)) == SOCKET_ERROR) {
    if (DEBUGGING) DB->Debug_Print("Error " + String(WSAGetLastError()) + " when getting local host name");
	WSACleanup();
    return -1;
    }
if (DEBUGGING) DB->Debug_Print("Host name is " + String(ac));

IPComboBox->Items->Clear();
struct hostent *phe = gethostbyname(ac);
if (phe == 0) {
    if (DEBUGGING) DB->Debug_Print("Bad host lookup");
    WSACleanup();
    return -1;
    }

for (int i = 0; phe->h_addr_list[i] != 0; ++i) {
    struct in_addr addr;
    memcpy(&addr, phe->h_addr_list[i], sizeof(struct in_addr));
    IPComboBox->Items->Append(inet_ntoa(addr));
    if (DEBUGGING) DB->Debug_Print("Network Found : " + String(inet_ntoa(addr)));
    IPCount++;
	}
    
WSACleanup();
// set a value into the Item Index if possible
if (IPComboBox->Items->Count > 0) {
	IPComboBox->ItemIndex = 0;
	} else {
	IPComboBox->ItemIndex = -1;
	}
return IPCount;
}

//---------------------------------------------------------------------------

void __fastcall TSetup::IPComboBoxChange(TObject *Sender)
{
// restart the mDNSServer
StartmDNSServices(RFID_WiFi_Mode->Checked, Radio_WiFi_Mode->Checked);	// Start the mDNS responder if in WiFi mode
}
//---------------------------------------------------------------------------

void __fastcall TSetup::ScanHostIPButtonClick(TObject *Sender)
{
int IPCount = FindIPs();
if (IPCount < 0) {
    MainForm->DebugMessagePopup("Warning", "A Network is Required for WiFi Operations");
	return;
    }
if (IPCount == 1) {
    IPComboBox->ItemIndex = 0;  // Select the first (and only) IP address
    StartmDNSServices(RFID_WiFi_Mode->Checked, Radio_WiFi_Mode->Checked);	// Start the mDNS responder if in WiFi mode
	} else {
    IPComboBox->ItemIndex = -1; // indicate that no IP has yet be selected
    MainForm->DebugMessagePopup("Host IP", "Select a Host IP for your Network - Use the 'Host IP Address' tab in Setup");
    }
}
//---------------------------------------------------------------------------

void __fastcall TSetup::NoSleepClick(TObject *Sender)
{
// Sleep
if (NoSleep->Checked) {
    // We could #include windows.h to get these constants
    #define ES_CONTINUOUS 0x80000000
    #define ES_SYSTEM_REQUIRED 0x00000001
    #define ES_AWAYMODE_REQUIRED 0x00000040
    // This sets the appropriate flags to prevent the system going into sleep mode; this is useful if making a screen saver that processes data in the background and can't have the computer sleep
    SetThreadExecutionState(ES_CONTINUOUS | ES_SYSTEM_REQUIRED | ES_AWAYMODE_REQUIRED);
    }
if (!NoSleep->Checked) {
    // We could #include windows.h to get this constant
    #define ES_CONTINUOUS 0x80000000
	// This allows the system to sleep normally
	SetThreadExecutionState(ES_CONTINUOUS);
	}
//
}
//---------------------------------------------------------------------------

