//---------------------------------------------------------------------------
/*
Remote Train Control Program for Windows

 Copyright 2015 by Mark DiVecchio

    This file is part of Remote Train Control.

    Remote Train Control is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    Remote Train Control is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Remote Train Control.  If not, see <http://www.gnu.org/licenses/>.

*/
// Home Page : http://www.silogic.com/trains/RTC_Running.html
//---------------------------------------------------------------------------
//---------------------------------------------------------------------------
#include <vcl.h>
#include <tchar.h>
#pragma hdrstop

#include "RTCMegaMaster.h"		// Dispatch List

#include "Message_Popup.h"
#include "Engine_Commands.h"
#include "Train_Control.h"
#include "Op_Control.h"
#include "Volume_Control.h"
#include "Track_Signal_Thread.h"
#include "Track_Signal.h"
#include "Debug_Controls.h"
#include "Setup_Control.h"
#include "Comm_Thread.h"
#include "perfgrap.h"

#include <math.h>
#include <stdio.h>
#include <string.h>

#pragma package(smart_init)


extern TCommThread * ptrTIUCommThread;				// pointer to TIU communications thread
extern TVolume * ptrVolume[TIUMAX][ABSOLUTEMAX];	// Volume Control Pointers - indexed by TIUNo-1 and actual engine number because 0 is a factory reset engine

//---------------------------------------------------------------------------
//   Important: Methods and properties of objects in VCL can only be
//   used in a method called using Synchronize, for example:
//
//      Synchronize(UpdateCaption);
//
//   where UpdateCaption could look like:
//
//      void __fastcall TTrackSignalThread::UpdateCaption()
//      {
//        Form1->Caption = "Updated in a thread";
//      }
//---------------------------------------------------------------------------
__fastcall TTrackSignalThread::TTrackSignalThread(bool CreateSuspended)
	: TThread(CreateSuspended)
{
assert(false);	// never called since I used the overloaded constructor below
}

//---------------------------------------------------------------------------
__fastcall TTrackSignalThread::TTrackSignalThread(bool CreateSuspended, TTrackSignal * CallingTS, HWND Parent,
		TOp * CallingOp, int Eng, int TIU, int Remote, DWORD delay, bool RepeatParameter)
	: TThread(CreateSuspended)
{
// Calling window
CallingTSScreen = CallingTS;
// Calling Operations window
CallingOpScreen = CallingOp;
// Engine Number and TIU Number
EngineNo = Eng;
TIUNo = TIU;
RemoteNo = Remote;
// Milliseconds to wait before repeating
DelayValue = delay;
// Always repeats so this parameter is not used
bRepeatParameter = RepeatParameter;
Finished = false;
// FreeOnTerminate is not set so caller must delete thread
// ReturnValue is accessable using WaitFor()
ReturnValue = -1;
//
// Parent Handle - right now, this value is not used
ParentHandle = Parent;
// termination
OnTerminate = &RunOnTerminate;

PingbyteResp[0] = NULL;
}
//---------------------------------------------------------------------------
__fastcall TTrackSignalThread::~TTrackSignalThread()
{
if (DEBUGGINGX(1)) {
	// print the ThreadId
    ptrTIUCommThread->PrintItem("TTrackSignalThread:~ThreadID " + IntToHex((int)ThreadID, 8), (TColor) clGreen);
    }
Finished = true;
}
//---------------------------------------------------------------------------

void __fastcall TTrackSignalThread::Execute()
{
// Calls to Sleep() are ok in this thread
// NO VCL Calls in this thread
if (Terminated) {
	return;						// in case we are terminated
    }
if (DEBUGGINGX(1)) {
	// print the ThreadId
    ThreadID = GetCurrentThreadId();
    ptrTIUCommThread->PrintItem("TTrackSignalThread: ThreadID " + IntToHex((int)ThreadID, 8), (TColor) clGreen);
    }
// this does not seem to work in BCB4.0 and Thread view not available in CB10.1
// it does work on BCB6.0
// http://docwiki.embarcadero.com/RADStudio/Rio/en/Converting_an_Unnamed_Thread_to_a_Named_Thread
THREADNAME_INFO info;
info.dwType = 0x1000;
info.szName = "Track_Signal_Thread";
info.dwThreadID = -1;
info.dwFlags = 0;
__try {
    RaiseException(0x406D1388, 0, sizeof(info) / sizeof(DWORD),
        (DWORD*)&info);
    }
__except (EXCEPTION_CONTINUE_EXECUTION) {
    }
//
Synchronize(Initialize);		// Perform the setup synchronously
//
while (true) {
    //
    if (Terminated) break;			// in case we are terminated
    //
	bool bretv = ptrTIUCommThread->Send_String_FIFO(EngineNo, "p4", TIUNo, RemoteNo, 15, LONG_WAIT, NO_RECORD, NULL, NULL, NULL, PingbyteResp);
    if (bretv) Synchronize(TS);		// Display the results so far
    if (Terminated) break;
    Sleep(DelayValue/5);			// DelayValue ticks split 5 ways
	if (Terminated) break;
    Sleep(DelayValue/5);			// DelayValue ticks split 5 ways
	if (Terminated) break;
    Sleep(DelayValue/5);			// DelayValue ticks split 5 ways
	if (Terminated) break;
    Sleep(DelayValue/5);			// DelayValue ticks split 5 ways
	if (Terminated) break;
    Sleep(DelayValue/5);			// DelayValue ticks split 5 ways
	if (Terminated) break;
	}	// while(true)
//
Synchronize(CleanUp);	// Perform the cleanup synchronously
ReturnValue = SUCCESS;
}
//---------------------------------------------------------------------------
static void __fastcall (__closure *VolumeOnValueChange)(TObject *Sender);
//---------------------------------------------------------------------------

void __fastcall TTrackSignalThread::Initialize(void)     // called using Synchronize()
{
/*
I've learned not to intermix calls to MainForm->Delay() and Sleep(). Sleep() is ok
in the Execute() function but this function runs as the main thread which
makes a lot of calls to send commands to the TIU which uses MainForm->Delay().
*/
if (CallingOpScreen->SoundsSwitch->Checked) {	// if sound is on, turn it off
	// Sound off - Remote turns off V1 (Engine) and V2 (Accent)
	(void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "v1000", TIUNo, RemoteNo, 11, SHORT_WAIT, NO_RECORD);
	(void)ptrTIUCommThread->Send_String_FIFO(EngineNo, "v2000", TIUNo, RemoteNo, 11, SHORT_WAIT, NO_RECORD);
    //
	TVolume * ptrVol = ptrVolume[TIUNo - 1][EngineNo];
	if (ptrVol) {	// reflect the change on to any open Sound Controls window
   		TAbVSlider* ptrV = ptrVol->EngineVolume;
	    VolumeOnValueChange = ptrV->OnValueChange;
   		ptrV->OnValueChanged = NULL;
	    ptrV->Value = 0;
		ptrV->OnValueChange = VolumeOnValueChange;
        //
   		ptrV = ptrVol->AccentVolume;
	    VolumeOnValueChange = ptrV->OnValueChange;
   		ptrV->OnValueChanged = NULL;
	    ptrV->Value = 0;
		ptrV->OnValueChange = VolumeOnValueChange;
	   	}
    }
//
p4count = 0;	// initialiaze counts and averages
p4ave = 0;
p4total = 0;
//
TestValue = 1;
}
//---------------------------------------------------------------------------

void __fastcall TTrackSignalThread::CleanUp(void)     // called using Synchronize()
{
/*
I've learned not to intermix calls to MainForm->Delay() and Sleep(). Sleep() is ok
in the Execute() function but this function runs as the main thread which
makes a lot of calls to send commands to the TIU which uses MainForm->Delay().
*/
if (CallingOpScreen->SoundsSwitch->Checked) { 	// if sound should be on, turn it on.
   	char str[15];
	// Make sure that correct engine data is loaded
	MainForm->ChangeTIUData(MainForm->EngineData, TIUNo);
	// Sound back on
    snprintf(str, sizeof str, "v1%03d", MainForm->EngineData->Cells[VOL_ENGINE][EngineNo].ToInt());
	(void)ptrTIUCommThread->Send_String_FIFO(EngineNo, str, TIUNo, RemoteNo, 11, SHORT_WAIT, NO_RECORD);
    snprintf(str, sizeof str, "v2%03d", MainForm->EngineData->Cells[VOL_ACCENT][EngineNo].ToInt());
   	(void)ptrTIUCommThread->Send_String_FIFO(EngineNo, str, TIUNo, RemoteNo, 11, SHORT_WAIT, NO_RECORD);
    //
    TVolume * ptrVol = ptrVolume[TIUNo - 1][EngineNo];
	if (ptrVol) {	// reflect the change on to any open Sound Controls window
   		TAbVSlider* ptrV = ptrVol->EngineVolume;
	    VolumeOnValueChange = ptrV->OnValueChange;
   		ptrV->OnValueChanged = NULL;
	    ptrV->Value = MainForm->EngineData->Cells[VOL_ENGINE][EngineNo].ToInt();
		ptrV->OnValueChange = VolumeOnValueChange;
        //
   		ptrV = ptrVol->AccentVolume;
	    VolumeOnValueChange = ptrV->OnValueChange;
   		ptrV->OnValueChanged = NULL;
	    ptrV->Value = MainForm->EngineData->Cells[VOL_ACCENT][EngineNo].ToInt();
		ptrV->OnValueChange = VolumeOnValueChange;
	   	}
    }
}
//---------------------------------------------------------------------------

void __fastcall TTrackSignalThread::TS(void)     // called using Synchronize()
{
/*

    DCS Signal Strength

    DCS signal strength is a number, from 1-10, that's calculated based on the
    number of data packets that the TIU sends and receives back from a PS2 engine
    that's performing a DCS signal strength test.


    During the test, the TIU sends a continuous series of data packets to the PS2
    engine. The PS2 engine is expected to reply to every packet that it receives,
    however, some packets are either not received by the engine or the response is
    not received by the TIU. Regardless, the TIU calculates DCS signal strength by
    counting how many response packets the TIU receives from the PS2 engine out of
    each hundred data packets the TIU sent, and then looks at a sliding scale to
    determine the signal strength that should be reported. The scale is not linear,
    rather, it is such that 87-100 packets equates to a DCS signal strength of 10,
    80-86 equates to a 9, and so on.


I've learned not to intermix calls to MainForm->Delay() and Sleep(). Sleep() is ok
in the Execute() function but this function runs as the main thread which
makes a lot of calls to send commands to the TIU which uses MainForm->Delay().
*/

// typical response to p4
//5555555A65A56569569A5559A9555A694D
//  00  05  70  34  47  04  A8  1E
//         'p' '4'
//55554D
//  00
//A59959AA5559555555565955
//  E4  5D  04  00  01  08
//                ^^^^ ------ Indicates there is 1 data byte in the response
//                    ^^^^ -- this is the response - the number of packets to the engine out of 100 that were lost.
//
// pull out the number of packets lost out of 100
const int p4loss = PingbyteResp[5];
if (p4loss > 100) {
    // invalid response value
    return;
	}
p4count++;
const int p4Val = (100 - p4loss);	// calculate the "X out of 100" score. We display the number of good packets out of 100
p4total += p4Val;
p4ave = p4total / p4count;
CallingTSScreen->FTS->Text =        "  " + String(p4Val) + " out of 100";

// River Nights     - https://www.heavy.ai/blog/12-color-palettes-for-telling-better-stories-with-your-data
TColor ColorHex[] = {0x0000b3, 0x58117c, 0xaf2144, 0xff531a, 0xe6880d, 0xc7b700, 0x5ad45a, 0x4ee08b, 0x78dceb};
// update graph with the colors of the rainbow
if (p4Val >= 87) {      // 87-100 = 10
    CallingTSScreen->Signals->DataPoint(ColorHex[7], p4Val);
    } else if (p4Val >= 80) {       // 80-86 = 9
        CallingTSScreen->Signals->DataPoint(ColorHex[6], p4Val);
        } else if (p4Val >= 75) {       // 8
            CallingTSScreen->Signals->DataPoint(ColorHex[5], p4Val);
            } else if (p4Val >= 68) {       // 7
                CallingTSScreen->Signals->DataPoint(ColorHex[4], p4Val);
                } else if (p4Val >= 61) {       // 6
                    CallingTSScreen->Signals->DataPoint(ColorHex[3], p4Val);
                    } else if (p4Val >= 54) {       // 5
                        CallingTSScreen->Signals->DataPoint(ColorHex[2], p4Val);
                        } else if (p4Val >= 47) {       //4
                            CallingTSScreen->Signals->DataPoint(ColorHex[1], p4Val);
                                } else {                        // 3
                                CallingTSScreen->Signals->DataPoint(ColorHex[0], p4Val);
                                }

CallingTSScreen->Signals->Update();

CallingTSScreen->FTSAverage->Text = "  " + String(p4ave) + " out of 100";

if (DEBUGGINGX(1)) {
   	DB->Debug_Print("Engine " + String(EngineNo) + " - TIU " + String(TIUNo) + " Instant: " + CallingTSScreen->FTS->Text);
	DB->Debug_Print("Engine " + String(EngineNo) + " - TIU " + String(TIUNo) + " Average: " + CallingTSScreen->FTSAverage->Text);
    }
//
DelayValue = CallingTSScreen->UpdateRateSpin->ValueAsInt * 1000L;
CallingTSScreen->QueryLabel->Caption = "Query # " + String(TestValue++);
}
//---------------------------------------------------------------------------

void __fastcall TTrackSignalThread::RunOnTerminate(TObject * Sender)     // called Automatically using Synchronize()
{
// not used
}
//---------------------------------------------------------------------------

